-- See the Copyright notice at the end of this file.
--
deferred class STREAM
	--
	-- A stream of characters.
	--
	-- There are two kinds of streams:
	-- + input streams (see INPUT_STREAM)
	-- + output_streams (see OUTPUT_STREAM)
	--
	-- Streams can:
	-- + be connected (e.g. to some system object)
	-- + be used ot read or write characters, only if they are connected
	-- + be filtered (see FILTER)
	--

insert
   ANY

feature {ANY}
	is_connected: BOOLEAN is
         -- True if the stream is connected. Only in that case can data be transfered via this stream.
		deferred
		end

	disconnect is
         -- Try to disconnect the stream. Note that it *does not* ensure that the stream will effectively be
         -- disconnected (some terminal streams, for instance, are always connected) but the feature can be
         -- used to "shake off" filters.
		require
			is_connected
         can_disconnect
		deferred
		ensure
			not is_filtered
		end

	is_filtered: BOOLEAN is
         -- True if some filter is using this stream as backend. use that filter instead.
		do
			Result := filter /= Void
		end

	detach is
         -- Shake off the filter.
		deferred
		ensure
			not is_filtered
		end

   descriptor: INTEGER is
         -- Some OS-dependent descriptor. Mainly used by the sequencer library (see READY_CONDITION).
      require
         is_connected
         has_descriptor
      do
         Result := filtered_descriptor
      end

   has_descriptor: BOOLEAN is
         -- True if that stream can be assocpiated to some OS-meaningful descriptor.
      require
         is_connected
      do
         Result := filtered_has_descriptor
      end

   can_disconnect: BOOLEAN is
         -- True if the stream can be safely disconnected (without data loss, etc.)
      require
         is_connected
      deferred
      end

feature {FILTER}
	set_filter (a_filter: like filter) is
         -- Used by the filter itself to get attached
		require
			a_filter /= Void
		do
			filter := a_filter
		ensure
			filter = a_filter
		end

	filter: FILTER
         -- The filter that uses this stream as backend

   filtered_descriptor: INTEGER is
         -- Find the descriptor of the terminal stream... Filters do not have descriptors of their own
      require
         is_connected
         filtered_has_descriptor
      deferred
      end

   filtered_has_descriptor: BOOLEAN is
         -- True if the underlying terminal stream has a descriptor
      require
         is_connected
      deferred
      end

feature {}
	basic_file_descriptor (file: POINTER): INTEGER is
		external "plug_in"
		alias "{
         location: "${sys}runtime"
         module_name: "basic_file"
         feature_name: "basic_file_descriptor"
         }"
		end

end -- class STREAM
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- This file is free software, which comes along with SmartEiffel. This software is distributed in the hope that it will be
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- You can modify it as you want, provided this footer is kept unaltered, and a notification of the changes is added.
-- You are allowed to redistribute it and sell it, alone or as a part of another product.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
