-- See the Copyright notice at the end of this file.
--
class GENERAL
	--
	-- Platform-independent universal properties. This class is an ancestor to all developer-written classes.
	--

feature {}
	default_create is
			-- Default creation method. It is used when no creation
			-- method is specified if allowed. Note it may be renamed.
		do
		end

	frozen manifest_creation is
			-- This procedure is used as a placeholder in creation
			-- clauses. Instances of a class can be created using the
			-- "manifest generic" notation if and only if this class has
			-- `manifest_creation' in its creation clause.
			-- You should never call this procedure.
		do
			check
				False
			end
		end

feature {ANY} -- Access:
	generating_type: STRING is
			-- Name of current object's generating type (type of
			-- which it is a direct instance).
		external "built_in"
		end

	generator: STRING is
			-- Name of current object's generating class (base class
			-- of the type of which it is a direct instance).
		external "built_in"
		end

feature {ANY} -- Status report:
	frozen same_dynamic_type (other: like Current): BOOLEAN is
			-- Is the dynamic type of `Current' identical to the dynamic type of `other'?
		require
			other /= Void
		external "built_in"
		end

feature {ANY} -- Comparison:
	frozen equal (some: ANY; other: like some): BOOLEAN is
			-- Are `some' and `other' both Void or attached to objects considered equal ?
		obsolete "Use `is_equal' instead (aug 28th 2004)."
		do
			if some = other then
				Result := True
			elseif some = Void then
			elseif other = Void then
			else
				Result := some.is_equal(other)
			end
		end

	is_equal (other: like Current): BOOLEAN is
			-- Is `other' attached to an object considered equal to
			-- current object ?
		require
			other /= Void
		external "built_in"
		ensure
			generating_type = other.generating_type implies Result = other.is_equal(Current)
		end

	frozen standard_equal (some: ANY; other: like some): BOOLEAN is
			-- Are `some' and `other' both Void or attached to
			-- field-by-field objects of the same type ?
			-- Always use the default object comparison criterion.
		obsolete "Use `standard_is_equal' instead (aug 28th 2004)."
		do
			if some = other then
				Result := True
			elseif some = Void then
			elseif other = Void then
			elseif some.same_dynamic_type(other) then
				Result := some.standard_is_equal(other)
			end
		end

	frozen standard_is_equal (other: like Current): BOOLEAN is
			-- Are Current and `other' field-by-field identical?
		require
			other /= Void
		external "built_in"
		ensure
			same_dynamic_type(other) implies Result = other.standard_is_equal(Current)
		end

feature {ANY} -- Deep Comparison:
	frozen deep_equal (some: ANY; other: like some): BOOLEAN is
			-- Are `some' and `other' either both Void or attached to
			-- recursively isomorphic object structures ?
		obsolete "Use `is_deep_equal' instead (aug 28th 2004)."
		do
			if some = other then
				Result := True
			elseif some = Void then
			elseif other = Void then
			else
				Result := some.is_deep_equal(other)
			end
		end

	is_deep_equal (other: like Current): BOOLEAN is
			-- Is `Current' recursively isomorph with `other' ?
		require
			other_not_void: other /= Void
		external "built_in"
		end

feature {ANY} -- Duplication:
	frozen clone (other: ANY): like other is
			-- When argument `other' is Void, return Void otherwise return `other.twin'.
		obsolete "Use `twin' instead (aug 28th 2004)."
		do
			if other /= Void then
				Result := other.twin
			end
		end

	frozen twin: like Current is
			-- Return a new object with the dynamic type of Current.
			-- Before being returned, the new object is initialized using feature `copy' (Current is passed as the argument).
			-- Thus, when feature `copy' of ANY is not redefined, `twin' has exactly the same behaviour as `standard_twin'.
		external "built_in"
		ensure
			equal: Result.is_equal(Current)
		end

	copy (other: like Current) is
			-- Update current object using fields of object attached
			-- to `other', so as to yield equal objects.
			-- Note: you can't copy object from a different dynamic type.
		require
			same_dynamic_type(other)
		external "built_in"
		ensure
			is_equal(other)
		end

	frozen standard_clone (other: ANY): like other is
			-- Void if `other' is Void; otherwise new object
			-- field-by-field identical to `other'.
			-- Always use the default copying semantics.
		obsolete "Use `standard_twin' instead (aug 28th 2004)."
		do
			if other /= Void then
				Result := other.standard_twin
			end
		end

	frozen standard_twin: like Current is
			-- Return a new object with the dynamic type of Current.
			-- Before being returned, the new object is initialized using
			-- feature `standard_copy' (Current is passed as the argument).
		external "built_in"
		end

	frozen standard_copy (other: like Current) is
			-- Copy every field of `other' onto corresponding field of
			-- current object.
		require
			other_not_void: other /= Void
		external "built_in"
		ensure
			standard_is_equal(other)
		end

feature {ANY} -- Deep Duplication:
	frozen deep_clone (other: ANY): like other is
			-- When argument `other' is Void, return Void otherwise return `other.deep_twin'.
		obsolete "Use `deep_twin' instead (aug 28th 2004)."
		do
			if other /= Void then
				Result := other.deep_twin
			end
		end

	deep_twin: like Current is
			-- Return a new object with the dynamic type of Current.
			-- The new object structure is recursively duplicated from the one
			-- attached to `Current'.
		external "built_in"
		end

feature {ANY} -- Basic operations:
	frozen default: like Current is
			-- Default value of entities declared with the `Current' type.
			-- Hence, the `Result' is Void for all reference types and, as another
			-- example, the `Result' is 0 for an INTEGER expression.
		do
		end

	is_default: BOOLEAN is
			-- Is the `Current' object in the `default' state?
			-- For example, when this feature is applied on an INTEGER, the
			-- `Result' is True only when the INTEGER is 0.  If not redefined, the
			-- `Result' is always False for a reference expression. Actually, this
			-- `is_default' predicate may be useful for user-defined expanded
			-- objects stored in collection (see `all_default' of class
			-- COLLECTION).
		do
			Result := Current = default
		end

feature{}
	frozen default_pointer: POINTER is
			-- Default value of type POINTER (avoid the need to write p.default
			-- for some `p' of type POINTER).
		do
		ensure
			Result = Result.default
		end

	default_rescue is
			-- Handle exception if no Rescue clause. (Default: do nothing.)
		do
		end

	frozen do_nothing is
			-- Execute a null action.
		do
		end

feature {} -- Input and Output:
	io: STD_INPUT_OUTPUT is
			-- Handle to standard file setup.
			-- To use the standard input/output file.
			-- Has type STD_FILES in ELKS 95.
		do
			Result := standard_streams.io
		ensure
			Result /= Void
		end

	std_input: INPUT_STREAM is
			-- To use the standard input file.
		do
			Result := standard_streams.std_input
		ensure
			Result /= Void
		end

	std_output: OUTPUT_STREAM is
			-- To use the standard output file.
		do
			Result := standard_streams.std_output
		ensure
			Result /= Void
		end

	std_error: OUTPUT_STREAM is
			-- To use the standard error file.
		do
			Result := standard_streams.std_error
		ensure
			Result /= Void
		end

	standard_streams: STANDARD_STREAMS is
		once
			create Result.make
		ensure
			Result /= Void
		end

feature {} -- Object Printing:
	frozen print (some: STRING) is
			-- When some is Void, output "Void" on `std_output', otherwise, output `some' on `std_output'.
			-- Note: still here for historal reasons.
		do
			if some = Void then
				std_output.put_string(once "Void")
			else
				std_output.put_string(some)
			end
		end

feature {ANY} -- Object Printing:
	print_on (file: OUTPUT_STREAM) is
			-- Default printing of current object on a `file'.
			-- One may redefine `fill_tagged_out_memory' or `out_in_tagged_out_memory' to adapt the behavior of
			-- `print_on'.
			--
		do
			tagged_out_memory.clear_count
			out_in_tagged_out_memory
			file.put_string(tagged_out_memory)
		end

	frozen tagged_out: STRING is
			-- New string containing printable representation of current
			-- object, each field preceded by its attribute name, a
			-- colon and a space.
		do
			tagged_out_memory.clear_count
			fill_tagged_out_memory
			Result := tagged_out_memory.twin
		end

	out: STRING is
			-- Create a new string containing terse printable
			-- representation of current object.
		do
			tagged_out_memory.clear_count
			out_in_tagged_out_memory
			Result := tagged_out_memory.twin
		end

	out_in_tagged_out_memory is
			-- Append terse printable represention of current object
			-- in `tagged_out_memory'.
		do
			tagged_out_memory.append(generating_type)
			if not is_expanded_type then
				tagged_out_memory.extend('#')
				to_pointer.append_in(tagged_out_memory)
			end
			tagged_out_memory.extend('[')
			fill_tagged_out_memory
			tagged_out_memory.extend(']')
		ensure
			not_cleared: tagged_out_memory.count >= old tagged_out_memory.count
			append_only: (old tagged_out_memory.twin).is_equal(tagged_out_memory.substring(1, old tagged_out_memory.count))
		end

	frozen tagged_out_memory: STRING is
		once
			create Result.make(1024)
		end

	fill_tagged_out_memory is
			-- Append a viewable information in `tagged_out_memory' in
			-- order to affect the behavior of `out', `tagged_out', etc.
		do
			-- Should be an external "built_in" to provide a default
			-- view of Current contents (not yet implemented).
		end

feature {} -- Basic named file handling:
	frozen file_tools: FILE_TOOLS is
		once
		end

	file_exists (path: STRING): BOOLEAN is
			-- True if `path' is an existing readable file.
		require
			not path.is_empty
		do
			Result := file_tools.is_readable(path)
		end

	remove_file (path: STRING) is
		require
			path /= Void
		do
			file_tools.delete(path)
		end

	rename_file (old_path, new_path: STRING) is
		require
			old_path /= Void
			new_path /= Void
		do
			file_tools.rename_to(old_path, new_path)
		end

feature {} -- Access to command-line arguments:
	argument_count: INTEGER is
			-- Number of arguments given to command that started
			-- system execution (command name does not count).
		do
			Result := command_arguments.upper
		ensure
			Result >= 0
		end

	argument (i: INTEGER): STRING is
			-- `i' th argument of command that started system execution
			-- Gives the command name if `i' is 0.
		require
			i >= 0
			i <= argument_count
		do
			Result := command_arguments.item(i)
		ensure
			Result /= Void
		end

	frozen command_arguments: FAST_ARRAY[STRING] is
			-- Give acces to arguments command line including the
			-- command name at index 0.
		local
			i: INTEGER; arg: STRING
		once
			from
				i := se_argc
				create Result.make(i)
			until
				i = 0
			loop
				i := i - 1
				arg := se_argv(i)
				Result.put(arg, i)
			end
		ensure
			not Result.is_empty
		end

feature {ANY} -- SCOOP:
	available: BOOLEAN is
			-- Wait the result (By Necessity [TM])
			-- Always return True
			-- (Meaningless in a non-SCOOP world)
		do
			Result := True
		end

feature {} -- Various useful tools:
	frozen crash is
			-- Print Run Time Stack and then exit with `exit_failure_code'.
		do
			print_all_run_time_stacks
			sedb_breakpoint
			die_with_code(exit_failure_code)
		end

	frozen trace_switch (flag: BOOLEAN) is
			-- May be used in combination with option "-sedb" of command
			-- `compile_to_c' (see compile_to_c documentation for details).
		external "built_in"
		end

	frozen sedb_breakpoint is
			-- May be used in combination with option "-sedb" of command
			-- `compile_to_c' to set a breakpoint for sedb, the SmartEiffel debugger.
		external "built_in"
		end

	frozen die_with_code (code: INTEGER) is
			-- Terminate execution with exit status code `code'.
			-- Do not print any message.
			-- Note: you can use predefined `exit_success_code' or
			-- `exit_failure_code' as well as another code you need.
		external "built_in"
		end

	exit_success_code: INTEGER is 0

	exit_failure_code: INTEGER is 1

	not_yet_implemented is
		do
			sedb_breakpoint
			std_error.put_string("[
            Sorry, some feature is not yet implemented (i.e. feature `not_yet_implemented' of
            class GENERAL has been called somewhere). If this is not the case, just run this
            code under the debugger to know the `not_yet_implemented' caller. If this is a
            feature of the SmartEiffel library, you may consider to post your implementation
            on the SmartEiffel mailing list.
                                                e-mail: SmartEiffel@loria.fr
                                          Happy debug and thanks in advance.
                                                 http://SmartEiffel.loria.fr
                                                       The SmartEiffel team.

            ]")
			crash
		end

feature {ANY} -- The Guru section:
	to_pointer: POINTER is
			-- This routine can be used only if the value of `Current' is really
			-- a memory address. (This is the case for all reference types and
			-- for the NATIVE_ARRAY type only.) Actually, this routine do
			-- nothing: the value of `Current' which is an address is returned
			-- unchanged. The compiler will emit a warning if you try to use
			-- `to_pointer' on some invalid type.
		external "built_in"
		end

	frozen is_expanded_type: BOOLEAN is
			-- Result is True if target static type is an expanded type.
			-- Useful for formal generic type.
		external "built_in"
		end

	frozen is_basic_expanded_type: BOOLEAN is
			-- The `Result' is True if the target static type is one of the following
			-- basic types: BOOLEAN, CHARACTER, INTEGER, REAL, DOUBLE or POINTER.
		external "built_in"
		ensure
			Result implies is_expanded_type
		end

	frozen object_size: INTEGER is
			-- Gives the size of the current object at first level
			-- only (pointed-to sub-object are not concerned).
			-- The result is given in number of CHARACTER.
		external "built_in"
		end

feature {} -- The Guru section:
	c_inline_h (c_code: STRING) is
			-- Target must be Current and `c_code' must be a manifest
			-- string. Write `c_code' in the heading C file.
		external "built_in"
		end

	c_inline_c (c_code: STRING) is
			-- Target must be Current and `c_code' must be a manifest
			-- string. Write `c_code' in the stream at current position.
		external "built_in"
		end

feature {} -- Implementation of GENERAL (do not use directly):
	print_run_time_stack is
			-- Prints the run time stack.
			-- The result depends both on compilation mode and
			-- target langage used (C or Java byte code).
			-- Usually, in mode -boost, no information is printed.
		external "built_in"
		end

	print_all_run_time_stacks is
			-- Prints all the run time stacks.
			-- The result depends both on compilation mode and
			-- target langage used (C or Java byte code).
			-- Usually, in mode -boost, no information is printed.
		external "built_in"
		end

	se_argc: INTEGER is
			-- To implement `command_arguments'
		external "built_in"
		end

	se_argv (i: INTEGER): STRING is
			-- To implement `command_arguments'
		external "built_in"
		end

	se_atexit is
			-- Called automatically at exit when `std_output' is used.
			-- (Do not call explicitely. Implementation purpose.)
		do
			std_output.flush
		end

end -- class GENERAL
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- This file is free software, which comes along with SmartEiffel. This software is distributed in the hope that it will be
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- You can modify it as you want, provided this footer is kept unaltered, and a notification of the changes is added.
-- You are allowed to redistribute it and sell it, alone or as a part of another product.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
