-- See the Copyright notice at the end of this file.
--
deferred class DICTIONARY[V, K]
	--
	-- Associative memory. Values of type `V' are stored using Keys of type `K'.
	--
	-- Well knowned implementations, see HASHED_DICTIONARY and AVL_DICTIONARY.
	-- 

inherit
	ANY
		redefine is_equal, copy
		end

feature {ANY} -- Counting:
	count: INTEGER is
			-- Actual `count' of stored elements.
		deferred
		end

	is_empty: BOOLEAN is
			-- Is it empty ?
		do
			Result := count = 0
		ensure
			Result = (count = 0)
		end

feature {ANY} -- Basic access:
	has (k: K): BOOLEAN is
			-- Is there a value currently associated with key `k'?
		require
			k /= Void
		deferred
		end

	at, infix "@" (k: K): V is
			-- Return the value associated to key `k'.
			-- (See also `reference_at' if V is a reference type.)
		require
			has(k)
		deferred
		end

	reference_at (k: K): V is
			-- Return Void or the value associated with key `k'. Actually, this
			-- feature is useful when the type of values (the type V) is a
			-- reference type, to avoid using `has' just followed by `at' to get
			-- the corresponding value.
		require
			k /= Void
			values_are_expanded: Result /= Void implies not Result.is_expanded_type
		deferred
		ensure
			has(k) implies Result = at(k)
		end

	fast_has (k: K): BOOLEAN is
			-- Is there a value currently associated with key `k'?
		require
			k /= Void
		deferred
		end

	fast_at (k: K): V is
			-- Return the value associated to key `k'.
			-- (See also `reference_at' if V is a reference type.)
		require
			fast_has(k)
		deferred
		end

	fast_reference_at (k: K): V is
			-- Return Void or the value associated with key `k'. Actually, this
			-- feature is useful when the type of values (the type V) is a
			-- reference type, to avoid using `has' just followed by `at' to get
			-- the corresponding value.
		require
			k /= Void
			values_are_expanded: Result /= Void implies not Result.is_expanded_type
		deferred
		ensure
			fast_has(k) implies Result = fast_at(k)
		end

feature {ANY}
	put (v: V; k: K) is
			-- Change some existing entry or `add' the new one. If there is
			-- as yet no key `k' in the dictionary, enter it with item `v'.
			-- Otherwise overwrite the item associated with key `k'.
		require
			k /= Void
		deferred
		ensure
			v = at(k)
		end

	add (v: V; k: K) is
			-- To add a new entry `k' with its associated value `v'. Actually,
			-- this is equivalent to call `put', but may run a little bit
			-- faster.
		require
			can_add(v, k)
		deferred
		ensure
			count = 1 + old count
			v = at(k)
		end

	can_add (v: V; k: K): BOOLEAN is
		do
			Result := not has(k)
		end

feature {ANY} -- Looking and searching some value:
	occurrences (v: V): INTEGER is
			-- Number of occurrences using `equal'.
			-- See also `fast_occurrences' to choose the appropriate one.
		local
			i: INTEGER; safe_equal: SAFE_EQUAL[V]
		do
			from
				i := 1
			until
				i > count
			loop
				if safe_equal.test(v, item(i)) then
					Result := Result + 1
				end
				i := i + 1
			end
		ensure
			Result >= 0
		end

	fast_occurrences (v: V): INTEGER is
			-- Number of occurrences using `='.
			-- See also `occurrences' to choose the appropriate one.
		local
			i: INTEGER
		do
			from
				i := 1
			until
				i > count
			loop
				if v = item(i) then
					Result := Result + 1
				end
				i := i + 1
			end
		ensure
			Result >= 0
		end

	key_at (v: V): K is
			-- Retrieve the key used for value `v' using `equal' for comparison.
		require
			occurrences(v) = 1
		local
			i: INTEGER; safe_equal: SAFE_EQUAL[V]
		do
			from
				i := 1
			until
				safe_equal.test(v, item(i))
			loop
				i := i + 1
			end
			Result := key(i)
		ensure
			(create {SAFE_EQUAL[V]}).test(at(Result), v)
		end

	fast_key_at (v: V): K is
			-- Retrieve the key used for value `v' using `=' for comparison.
		require
			fast_occurrences(v) = 1
		local
			i: INTEGER
		do
			from
				i := 1
			until
				v = item(i)
			loop
				i := i + 1
			end
			Result := key(i)
		ensure
			at(Result) = v
		end

feature {ANY} -- Removing:
	remove (k: K) is
			-- Remove entry `k' (which may exist or not before this call).
		require
			k /= Void
		deferred
		ensure
			not has(k)
		end

	clear is
		obsolete "You must decide to use `clear_count' or `clear_count_and_capacity' (may 9th 2004)."
		do
			clear_count
		ensure
			is_empty
		end

	clear_count is
			-- Discard all items (`is_empty' is True after that call). The internal `capacity' is not changed
			-- by this call. See also `clear_count_and_capacity' to select the most appropriate.
		deferred
		ensure
			is_empty: count = 0
			capacity = old capacity
		end

	clear_count_and_capacity is
			-- Discard all items (`is_empty' is True after that call). The internal `capacity' may also be
			-- reduced after this call. See also `clear_count' to select the most appropriate.
		deferred
		ensure
			is_empty: count = 0
			capacity <= old capacity
		end

	capacity: INTEGER is
			-- Approximation of the actual internal storage `capacity'. The `capacity' will grow automatically
			-- when needed (i.e. `capacity' is not a limit for the number of values stored). Also note that
			-- the `capacity' value may not be always accurate depending of the implementation (anyway, this
			-- `capacity' value is at least equals to `count').
		deferred
		end

feature {ANY} -- To provide iterating facilities:
	lower: INTEGER is 1

	upper: INTEGER is
		do
			Result := count
		ensure
			Result = count
		end

	valid_index (index: INTEGER): BOOLEAN is
		do
			Result := 1 <= index and then index <= count
		ensure
			Result = index.in_range(lower, upper)
		end

	item (index: INTEGER): V is
		require
			valid_index(index)
		deferred
		ensure
			Result = at(key(index))
		end

	key (index: INTEGER): K is
		require
			valid_index(index)
		deferred
		ensure
			at(Result) = item(index)
		end

	get_new_iterator_on_items: ITERATOR[V] is
		do
			create {ITERATOR_ON_DICTIONARY_ITEMS[V, K]} Result.make(Current)
		ensure
			Result /= Void
		end

	get_new_iterator_on_keys: ITERATOR[K] is
		deferred
		ensure
			Result /= Void
		end

	key_map_in (buffer: COLLECTION[K]) is
			-- Append in `buffer', all available keys (this may be useful to
			-- speed up the traversal).
		require
			buffer /= Void
		local
			i: INTEGER
		do
			from
				i := count
			until
				i < lower
			loop
				buffer.add_last(key(i))
				i := i - 1
			end
		ensure
			buffer.count = count + old buffer.count
		end

	item_map_in (buffer: COLLECTION[V]) is
			-- Append in `buffer', all available items (this may be useful to
			-- speed up the traversal).
		require
			buffer /= Void
		local
			i: INTEGER
		do
			from
				i := count
			until
				i < lower
			loop
				buffer.add_last(item(i))
				i := i - 1
			end
		ensure
			buffer.count = count + old buffer.count
		end

feature {ANY}
	is_equal (other: like Current): BOOLEAN is
			-- Do both dictionaries have the same set of associations?
			-- Keys are compared with `is_equal' and values are comnpared
			-- with the basic = operator. See also `is_equal_map'.
		local
			i: INTEGER
		do
			if Current = other then
				Result := True
			elseif count = other.count then
				from
					Result := True
					i := 1
				until
					not Result or else i > count
				loop
					if other.has(key(i)) then
						if other.at(key(i)) /= item(i) then
							Result := False
						else
							i := i + 1
						end
					else
						Result := False
					end
				end
			end
		ensure then
			Result implies count = other.count
		end

	is_equal_map (other: like Current): BOOLEAN is
			-- Do both dictionaries have the same set of associations?
			-- Both keys and values are compared with `is_equal'. See also `is_equal'.
		local
			i: INTEGER; k: K; safe_equal: SAFE_EQUAL[V]
		do
			if Current = other then
				Result := True
			elseif count = other.count then
				from
					Result := True
					i := 1
				until
					not Result or else i > count
				loop
					k := key(i)
					if other.has(k) then
						if not safe_equal.test(other.at(k), item(i)) then
							Result := False
						else
							i := i + 1
						end
					else
						Result := False
					end
				end
			end
		end

	copy (other: like Current) is
			-- Reinitialize by copying all associations of `other'.
		local
			i: INTEGER
		do
			clear_count
			from
				i := 1
			until
				i > other.count
			loop
				put(other.item(i), other.key(i))
				i := i + 1
			end
		end

feature {ANY} -- Agents based features:
	do_all (action: ROUTINE[ANY, TUPLE[V, K]]) is
			-- Apply `action' to every [V,K] associations of `Current'.
		local
			i: INTEGER; v: V; k: K
		do
			from
				i := lower
			until
				i > upper
			loop
				v := item(i)
				k := key(i)
				action.call([v, k])
				i := i + 1
			end
		end

	for_all (test: PREDICATE[ANY, TUPLE[V, K]]): BOOLEAN is
			-- Do all [V,K] associations satisfy `test'?
		local
			i: INTEGER; v: V; k: K
		do
			from
				Result := True
				i := lower
			until
				not Result or else i > upper
			loop
				v := item(i)
				k := key(i)
				Result := test.item([v, k])
				i := i + 1
			end
		end

	exists (test: PREDICATE[ANY, TUPLE[V, K]]): BOOLEAN is
			-- Does at least one [V,K] association satisfy `test'?
		local
			i: INTEGER; v: V; k: K
		do
			from
				i := lower
			until
				Result or else i > upper
			loop
				v := item(i)
				k := key(i)
				Result := test.item([v, k])
				i := i + 1
			end
		end

feature {ANY} -- Other features:
	internal_key (k: K): K is
			-- Retrieve the internal key object which correspond to the existing
			-- entry `k' (the one memorized into the `Current' dictionary).
		require
			has(k)
		deferred
		ensure
			Result.is_equal(k)
		end

feature {}
	make is
			-- Creates an empty dictionary.
		deferred
		ensure
			is_empty
		end

feature {}
	-- Implement manifest generic creation.

	manifest_make (needed_capacity: INTEGER) is
			-- Manifest creation of a dictionary.
		do
			make
		end
	
	manifest_put (index: INTEGER; v: V; k: K) is
		do
			put(v, k)
		end

	manifest_semicolon_check: INTEGER is 2
			-- Put semicolons between successive value-key pairs.
	
feature {}
	key_safe_equal (k1, k2: K): BOOLEAN is
			-- Because keys are never Void, we do not rely on the SAFE_EQUAL class.
		require
			k1 /= Void
			k2 /= Void
		do
			if k1 = k2 then
				Result := True
			elseif k1.same_dynamic_type(k2) then
				Result := k1.is_equal(k2)
			end
		end
	
invariant
	capacity >= count

end -- class DICTIONARY
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- This file is free software, which comes along with SmartEiffel. This software is distributed in the hope that it will be
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- You can modify it as you want, provided this footer is kept unaltered, and a notification of the changes is added.
-- You are allowed to redistribute it and sell it, alone or as a part of another product.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
