-- See the Copyright notice at the end of this file.
--
class AVL_DICTIONARY[V, K -> COMPARABLE]
	--
	-- Associative memory.
	-- Values of type `V' are stored using Keys of type `K'.
	--

inherit
	SIMPLE_DICTIONARY[V, K]
		redefine occurrences, fast_occurrences, key_at, fast_key_at
		end
	AVL_TREE[K]
		redefine
			root
		end

creation {ANY}
	make, manifest_creation

feature {ANY} -- Basic access:
	count: INTEGER

	capacity: INTEGER is
		do
			Result := count --*** + lost_nodes.capacity
		end

	infix "@", at (k: K): V is
			-- Return the value associated to key `k'.
			-- (See also `reference_at' if V is a reference type.)
		do
			Result := root.at(k).value
		end

	fast_at (k: K): V is
		do
			Result := root.fast_at(k).value
		end

	reference_at (k: K): V is
			-- Return Void or the value associated with key `k'. Actually, this
			-- feature is useful when the type of values (the type V) is a
			-- reference type, to avoid using `has' just followed by `at' to get
			-- the corresponding value.
		local
			n: AVL_DICTIONARY_NODE[V, K]
		do
			if root /= Void then
				n := root.at(k)
				if n /= Void then
					Result := n.value
				end
			end
		end

	fast_reference_at (k: K): V is
			-- Return Void or the value associated with key `k'. Actually, this
			-- feature is useful when the type of values (the type V) is a
			-- reference type, to avoid using `has' just followed by `at' to get
			-- the corresponding value.
		local
			n: AVL_DICTIONARY_NODE[V, K]
		do
			if root /= Void then
				n := root.fast_at(k)
				if n /= Void then
					Result := n.value
				end
			end
		end

feature {ANY}
	put, add (v: V; k: K) is
		do
			root := do_insert(root, v, k)			
		end

feature {ANY} -- Looking and searching some value:
	occurrences (v: V): INTEGER is
			-- Number of occurrences using `equal'.
			-- See also `fast_occurrences' to chose the apropriate one.
		do
			if root /= Void then
				Result := root.occurrences(v)
			end
		end

	fast_occurrences (v: V): INTEGER is
			-- Number of occurrences using `='.
			-- See also `occurrences' to chose the apropriate one.
		do
			if root /= Void then
				Result := root.fast_occurrences(v)
			end
		end

	key_at (v: V): K is
			-- Retrieve the key used for value `v' using `equal' for comparison.
		do
			Result := root.key_at(v)
		end

	fast_key_at (v: V): K is
			-- Retrieve the key used for value `v' using `=' for comparison.
		do
			Result := root.fast_key_at(v)
		end

feature {ANY} -- Removing:
	clear_count, clear_count_and_capacity is
			-- Discard all items.
		do
			if not is_empty then
				clear_nodes(root)
				root := Void
				count := 0
				map_dirty := True
			end
		end

feature {ANY} -- To provide iterating facilities:
	item (index: INTEGER): V is
			-- Return the item indexed by `index'.
		do
			if map_dirty then
				build_map
			end
			Result := map.item(index - 1).value
		end

	key (index: INTEGER): K is
		do
			if map_dirty then
				build_map
			end
			Result := map.item(index - 1).key
		end

	get_new_iterator_on_keys: ITERATOR[K] is
		do
			create {ITERATOR_ON_DICTIONARY_KEYS[V, K]} Result.make(Current)
		end

feature {ANY} -- Other features:
	internal_key (k: K): K is
			-- Retrieve the internal key object which correspond to the existing
			-- entry `k' (the one memorized into the `Current' dictionary).
		do
			Result := root.at(k).key
		end

feature {}
	root: AVL_DICTIONARY_NODE[V, K]

	do_insert (n: like a_new_node; v: V; k: K): like a_new_node is
		do
			if n = Void then
				Result := new_node
				Result.make(v, k)
				count := count + 1
				map_dirty := True
				rebalance := True
			elseif k < n.key then
				n.set_left(do_insert(n.left, v, k))
				if rebalance then
					Result := left_grown(n)
				else
					Result := n
				end
			elseif k > n.key then
				n.set_right(do_insert(n.right, v, k))
				if rebalance then
					Result := right_grown(n)
				else
					Result := n
				end
			else
				n.set_value(v)
				Result := n
				rebalance := False
			end
		ensure
			Result /= Void
			Result.balance = node_height(Result.right) - node_height(Result.left)
			rebalance = (node_height(Result) > old node_height(n))
		end

	exchange_and_discard (n1, n2: like root) is
		do
			n1.set_key(n2.key)
			n1.set_value(n2.value)
			rebalance := True
			count := count - 1
			discard_node(n2)
		end

	new_node: like a_new_node is
		do
			Result := lost_nodes.item
			if Result = Void then
				Result := a_new_node
			else
				lost_nodes.set_item(Result.left)
			end
		end

	discard_node (n: like a_new_node) is
		local
			v: V; k: K
		do
			n.make(v, k)
			n.set_left(lost_nodes.item)
			lost_nodes.set_item(n)
		end

	lost_nodes: WEAK_REFERENCE[like a_new_node]

	common_lost_nodes: DICTIONARY[WEAK_REFERENCE[ANY_AVL_DICTIONARY_NODE], STRING] is
		once
		 	create {HASHED_DICTIONARY[WEAK_REFERENCE[ANY_AVL_DICTIONARY_NODE], STRING]} Result.make
		end

	a_new_node: AVL_DICTIONARY_NODE[V, K] is
		do
			create Result
		end

feature {}
	make is
		do
			create map.make(0)
			lost_nodes ::= common_lost_nodes.reference_at(generating_type)
			if lost_nodes = Void then
				create lost_nodes.set_item(Void)
				common_lost_nodes.add(lost_nodes, generating_type)
			end
		end

invariant
	lost_nodes /= Void
	lost_nodes = common_lost_nodes.at(generating_type)

end -- class AVL_DICTIONARY
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- This file is free software, which comes along with SmartEiffel. This software is distributed in the hope that it will be
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- You can modify it as you want, provided this footer is kept unaltered, and a notification of the changes is added.
-- You are allowed to redistribute it and sell it, alone or as a part of another product.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
