-- See the Copyright notice at the end of this file.
--
deferred class DOUBLE_DICTIONARY[V, K]
	--
	-- Double dictionaries (symmetry value / key)
	-- (class akin to SIMPLE_DICTIONARY)
	--

inherit
	DICTIONARY[V, K]
		redefine can_add, clear, is_equal, occurrences, fast_occurrences, key_at, fast_key_at
		end

feature {ANY} -- Basic access:
	count: INTEGER is
		do
			Result := right.count
      end

   capacity: INTEGER is
      do
         Result := right.capacity
      end

	has (k: K): BOOLEAN is
			-- Is there a value currently associated with key `k'?
		do
			Result := right.has(k)
		end

	at, infix "@" (k: K): V is
			-- Return the value associated to key `k'.
			-- (See also `reference_at' if V is a reference type.)
		do
			Result := right.at(k)
		end

	reference_at (k: K): V is
			-- Return Void or the value associated with key `k'. Actually, this
			-- feature is useful when the type of values (the type V) is a
			-- reference type, to avoid using `has' just followed by `at' to get
			-- the corresponding value.
		do
			Result := right.reference_at(k)
		end

	fast_has (k: K): BOOLEAN is
			-- Is there a value currently associated with key `k'?
		do
			Result := right.fast_has(k)
		end

	fast_at (k: K): V is
			-- Return the value associated to key `k'.
			-- (See also `reference_at' if V is a reference type.)
		do
			Result := right.fast_at(k)
		end

	fast_reference_at (k: K): V is
			-- Return Void or the value associated with key `k'. Actually, this
			-- feature is useful when the type of values (the type V) is a
			-- reference type, to avoid using `has' just followed by `at' to get
			-- the corresponding value.
		do
			Result := right.fast_reference_at(k)
		end

feature {ANY}
	put (v: V; k: K) is
			-- Change some existing entry or `add' the new one. If there is as
			-- yet no key `k' nor item `v' in the dictionary, enter that new
			-- pair.  Otherwise overwrite the either item associated with key `k'
			-- or the key associated with the item `v'.
		do
			if right.occurrences(v) > 0 then
				right.remove(right.key_at(v))
			end
			if reverse.occurrences(k) > 0 then
				reverse.remove(reverse.key_at(k))
			end
			right.put(v, k)
			reverse.put(k, v)
		end

	add (v: V; k: K) is
			-- To add a new entry `k' with its associated value `v'. Actually,
			-- this is equivalent to call `put', but may run a little bit faster.
		do
			right.add(v, k)
			reverse.add(k, v)
		end

	can_add (v: V; k: K): BOOLEAN is
			-- Special case for double dictionaries: in order to add an entry,
			-- neither the key `k' nor the value `v' must be in the dictionary.
		do
			Result := not right.has(k) and then not reverse.has(v)
		end

feature {ANY} -- Looking and searching some value:
	occurrences (v: V): INTEGER is
			-- Number of occurrences using `equal'.
			-- See also `fast_occurrences' to chose the apropriate one.
		do
			if reverse.has(v) then
				Result := 1
			end
		ensure then
			Result = right.occurrences(v)
		end

	fast_occurrences (v: V): INTEGER is
			-- Number of occurrences using `='.
			-- See also `occurrences' to chose the apropriate one.
		do
			if reverse.fast_has(v) then
				Result := 1
			end
		ensure then
			Result = right.fast_occurrences(v)
		end

	key_at (v: V): K is
			-- Retrieve the key used for value `v' using `equal' for comparison.
		do
			Result := reverse.at(v)
		end

	fast_key_at (v: V): K is
			-- Retrieve the key used for value `v' using `=' for comparison.
		do
			Result := reverse.fast_at(v)
		end

feature {ANY} -- Removing:
	remove (k: K) is
			-- Remove entry `k' (which may exist or not before this call).
		local
			v: V
		do
			if has(k) then
				v := at(k)
				right.remove(k)
				reverse.remove(v)
			end
		end

	clear is
		obsolete "You must decide to use `clear_count' or `clear_count_and_capacity' (may 9th 2004)."
		do
			right.clear
			reverse.clear
		end

	clear_count is
		do
			right.clear_count
			reverse.clear_count
		end

	clear_count_and_capacity is
		do
			right.clear_count_and_capacity
			reverse.clear_count_and_capacity
		end

feature {ANY} -- To provide iterating facilities:
	item (index: INTEGER): V is
		do
			Result := right.item(index)
		end

	key (index: INTEGER): K is
		do
			Result := right.key(index)
		end

	get_new_iterator_on_keys: ITERATOR[K] is
		do
			Result := right.get_new_iterator_on_keys
		end

feature {ANY} -- Other features:
	is_equal (other: like Current): BOOLEAN is
		do
			not_yet_implemented
		end

	internal_key (k: K): K is
			-- Retrieve the internal key object which correspond to the existing
			-- entry `k' (the one memorized into the `Current' dictionary).
		do
			Result := right.internal_key(k)
		end

feature {}
	right: DICTIONARY[V, K]

	reverse: DICTIONARY[K, V]

invariant
   right.count = reverse.count
   right.capacity = reverse.capacity

end -- class DOUBLE_DICTIONARY
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- This file is free software, which comes along with SmartEiffel. This software is distributed in the hope that it will be
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- You can modify it as you want, provided this footer is kept unaltered, and a notification of the changes is added.
-- You are allowed to redistribute it and sell it, alone or as a part of another product.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
