-- See the Copyright notice at the end of this file.
--
class AVL_DICTIONARY_NODE[V, K -> COMPARABLE]
	--
	-- Auxiliary class to implement AVL_DICTIONARY[E].
	--

inherit
	AVL_TREE_NODE[K]
		rename item as key,
			set_item as set_key
		end
	ANY_AVL_DICTIONARY_NODE

creation {AVL_DICTIONARY}
	default_create

feature {AVL_DICTIONARY, AVL_DICTIONARY_NODE}
	value: V

	set_value (v: like value) is
		do
			value := v
		ensure
			value = v
		end

	fast_at (k: like key): AVL_DICTIONARY_NODE[V, K] is
			-- Is element `e' in the tree?
		local
			safe_equal: SAFE_EQUAL[K]
		do
			if key = k then
				Result := Current
			elseif safe_equal.test(key, k) then
				-- because otherwise there would be an infinite recursion
				--Result := Void
			elseif k < key then
				if left /= Void then
					Result := left.fast_at(k)
				end
			else
				if right /= Void then
					Result := right.fast_at(k)
				end
			end
		end

	occurrences (v: V): INTEGER is
		do
			Result := occurrences_(v, 0)
		end

	fast_occurrences (v: V): INTEGER is
		do
			Result := fast_occurrences_(v, 0)
		end

	key_at (v: V): K is
		local
			safe_equal: SAFE_EQUAL[V]
		do
			if safe_equal.test(v, value) then
				Result := key
			elseif left /= Void then
				Result := left.key_at(v)
			elseif right /= Void then
				Result := right.key_at(v)
			end
		end

	fast_key_at (v: V): K is
		do
			if v = value then
				Result := key
			elseif left /= Void then
				Result := left.fast_key_at(v)
			elseif right /= Void then
				Result := right.fast_key_at(v)
			end
		end

feature {AVL_DICTIONARY_NODE}
	occurrences_ (v: V; cnt: INTEGER): INTEGER is
		local
			safe_equal: SAFE_EQUAL[V]
		do
			Result := cnt
			if safe_equal.test(v, value) then
				Result := Result + 1
			end
			if left /= Void then
				Result := left.occurrences_(v, Result)
			end
			if right /= Void then
				Result := right.occurrences_(v, Result)
			end
		ensure
			Result >= cnt
		end

	fast_occurrences_ (v: V; cnt: INTEGER): INTEGER is
		do
			Result := cnt
			if v = value then
				Result := Result + 1
			end
			if left /= Void then
				Result := left.fast_occurrences_(v, Result)
			end
			if right /= Void then
				Result := right.fast_occurrences_(v, Result)
			end
		ensure
			Result >= cnt
		end

feature {AVL_DICTIONARY} -- Creation:
	make (v: like value; k: like key) is
		do
			set_balance(balanced)
			left := Void
			right := Void
			set_value(v)
			set_key(k)
		ensure
			value = v
			key = k
		end

end -- class AVL_DICTIONARY_NODE
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- This file is free software, which comes along with SmartEiffel. This software is distributed in the hope that it will be
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- You can modify it as you want, provided this footer is kept unaltered, and a notification of the changes is added.
-- You are allowed to redistribute it and sell it, alone or as a part of another product.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
