-- See the Copyright notice at the end of this file.
--
deferred class AVL_TREE[E -> COMPARABLE]
	--
	-- Definition of a mathematical set of comparable objects. All common
	-- operations on mathematical sets are available.
	--

insert
	AVL_CONSTANTS

feature {ANY}
	debug_string: STRING is
		do
			if root = Void then
				Result := once "Void"
			else
				Result := tagged_out_memory
				Result.clear_count
				root.out_in_tagged_out_memory
			end
		end

	count: INTEGER is
		deferred
		end

feature {ANY} -- Adding and removing:
	remove (e: E) is
			-- Remove item `e' from the set: the mathematical definition of
			-- removing from a set is followed.
		do
			root := do_remove(root, e)
		end

feature {}
	root: AVL_TREE_NODE[E]

	rebalance: BOOLEAN

	left_grown (n: like root): like root is
		require
			rebalance
			n /= Void
			node_height(n.right) - node_height(n.left) + 1 = n.balance
		do
			inspect n.balance
			when imbalanced_left then
				if n.left.balance = imbalanced_left then
					n.set_balance(balanced)
					n.left.set_balance(balanced)
				else
					inspect
						n.left.right.balance
					when imbalanced_left then
						n.set_balance(imbalanced_right)
						n.left.set_balance(balanced)
					when balanced then
						n.set_balance(balanced)
						n.left.set_balance(balanced)
					when imbalanced_right then
						n.set_balance(balanced)
						n.left.set_balance(imbalanced_left)
					end
					n.left.right.set_balance(balanced)
					n.set_left(n.left.rotate_left)
				end
				Result := n.rotate_right
				rebalance := False
			when balanced then
				n.set_balance(imbalanced_left)
				Result := n
			when imbalanced_right then
				n.set_balance(balanced)
				Result := n
				rebalance := False
			end
		ensure
			Result /= Void
			Result.balance = node_height(Result.right) - node_height(Result.left)
			rebalance = (node_height(Result) > 1 + old node_height(n.right).max(node_height(n.left) - 1))
		end

	right_grown (n: like root): like root is
		require
			rebalance
			n /= Void
			node_height(n.right) - 1 - node_height(n.left) = n.balance
		do
			inspect n.balance
			when imbalanced_right then
				if n.right.balance = imbalanced_right then
					n.set_balance(balanced)
					n.right.set_balance(balanced)
				else
					inspect
						n.right.left.balance
					when imbalanced_right then
						n.set_balance(imbalanced_left)
						n.right.set_balance(balanced)
					when balanced then
						n.set_balance(balanced)
						n.right.set_balance(balanced)
					when imbalanced_left then
						n.set_balance(balanced)
						n.right.set_balance(imbalanced_right)
					end
					n.right.left.set_balance(balanced)
					n.set_right(n.right.rotate_right)
				end
				Result := n.rotate_left
				rebalance := False
			when balanced then
				n.set_balance(imbalanced_right)
				Result := n
			when imbalanced_left then
				n.set_balance(balanced)
				Result := n
				rebalance := False
			end
		ensure
			Result /= Void
			Result.balance = node_height(Result.right) - node_height(Result.left)
			rebalance = (node_height(Result) > 1 + old node_height(n.left).max(node_height(n.right) - 1))
		end

	do_remove (n: like root; e: E): like root is
		do
			if n = Void then
				rebalance := False
			elseif e < n.item then
				n.set_left(do_remove(n.left, e))
				if rebalance then
					Result := left_shrunk(n)
				else
					Result := n
				end
			elseif e > n.item then
				n.set_right(do_remove(n.right, e))
				if rebalance then
					Result := right_shrunk(n)
				else
					Result := n
				end
			elseif n.left = Void then
				Result := n.right
				exchange_and_discard(n, n)
			elseif n.right = Void then
				Result := n.left
				exchange_and_discard(n, n)
			else
				n.set_left(remove_right(n, n.left))
				if rebalance then
					Result := left_shrunk(n)
				else
					Result := n
				end
			end
		ensure
			Result = Void or else Result.balance = node_height(Result.right) - node_height(Result.left)
			rebalance = (node_height(Result) < old node_height(n))
		end

	remove_right (n1, n2: like root): like root is
		require
			n1 /= Void
			n2 /= Void
		do
			if n2.right = Void then
				Result := n2.left
				exchange_and_discard(n1, n2)
			else
				n2.set_right(remove_right(n1, n2.right))
				if rebalance then
					Result := right_shrunk(n2)
				else
					Result := n2
				end
			end
		ensure
			Result = Void or else Result.balance = node_height(Result.right) - node_height(Result.left)
			rebalance = (node_height(Result) < old node_height(n2))
		end

	left_shrunk (n: like root): like root is
		require
			rebalance
			n /= Void
			node_height(n.right) - node_height(n.left) - 1 = n.balance
		do
			inspect
				n.balance
			when imbalanced_left then
				n.set_balance(balanced)
				Result := n
			when balanced then
				n.set_balance(imbalanced_right)
				Result := n
				rebalance := False
			when imbalanced_right then
				inspect
					n.right.balance
				when imbalanced_left then
					inspect
						n.right.left.balance
					when imbalanced_left then
						n.set_balance(balanced)
						n.right.set_balance(imbalanced_right)
					when balanced then
						n.set_balance(balanced)
						n.right.set_balance(balanced)
					when imbalanced_right then
						n.set_balance(imbalanced_left)
						n.right.set_balance(balanced)
					end
					n.right.left.set_balance(balanced)
					n.set_right(n.right.rotate_right)
				when balanced then
					n.set_balance(imbalanced_right)
					n.right.set_balance(imbalanced_left)
					rebalance := False
				when imbalanced_right then
					n.set_balance(balanced)
					n.right.set_balance(balanced)
				end
				Result := n.rotate_left
			end
		ensure
			Result /= Void
			Result.balance = node_height(Result.right) - node_height(Result.left)
			rebalance = (node_height(Result) < 1 + old node_height(n.right).max(node_height(n.left) + 1))
		end

	right_shrunk (n: like root): like root is
		require
			rebalance
			n /= Void
			node_height(n.right) + 1 - node_height(n.left) = n.balance
		do
			inspect
				n.balance
			when imbalanced_right then
				n.set_balance(balanced)
				Result := n
			when balanced then
				n.set_balance(imbalanced_left)
				Result := n
				rebalance := False
			when imbalanced_left then
				inspect
					n.left.balance
				when imbalanced_right then
					inspect
						n.left.right.balance
					when imbalanced_right then
						n.set_balance(balanced)
						n.left.set_balance(imbalanced_left)
					when balanced then
						n.set_balance(balanced)
						n.left.set_balance(balanced)
					when imbalanced_left then
						n.set_balance(imbalanced_right)
						n.left.set_balance(balanced)
					end
					n.left.right.set_balance(balanced)
					n.set_left(n.left.rotate_left)
				when balanced then
					n.set_balance(imbalanced_left)
					n.left.set_balance(imbalanced_right)
					rebalance := False
				when imbalanced_left then
					n.set_balance(balanced)
					n.left.set_balance(balanced)
				end
				Result := n.rotate_right
			end
		ensure
			Result /= Void
			Result.balance = node_height(Result.right) - node_height(Result.left)
			rebalance = (node_height(Result) < 1 + old node_height(n.left).max(node_height(n.right) + 1))
		end

	exchange_and_discard (n1, n2: like root) is
		require
			n1 /= Void
			n2 /= Void
		deferred
		ensure
			map_dirty
			count = old count - 1
			rebalance
		end

	clear_nodes (node: like root) is
		do
			if node.left /= Void then
				clear_nodes(node.left)
			end
			if node.right /= Void then
				clear_nodes(node.right)
			end
			discard_node(node)
		end

	node_height (node: like root): INTEGER is
		do
			if node /= Void then
				Result := node.height
			end
		end

feature {ANY} -- Looking and searching:
	has (e: E): BOOLEAN is
			-- Is element `e' in the set?
		do
			Result := root /= Void and then root.has(e)
		end

	fast_has (e: E): BOOLEAN is
			-- Is element `e' in the set?
		do
			Result := root /= Void and then root.fast_has(e)
		end

feature {} -- Iterating internals:
	build_map is
		require
			build_needed: map_dirty
		do
			map.clear_count
			if count > 0 then
				root.map_in(map)
			end
			map_dirty := False
		ensure
			build_done: not map_dirty
		end

	map: FAST_ARRAY[like root]
			-- Elements in a row for iteration. See `build_map'.

	map_dirty: BOOLEAN
			-- True when the map needs to be built again for the iterators. See
			-- `build_map'.

feature {}
	discard_node (n: like root) is
		require
			n /= Void
		deferred
		end

invariant
	map /= Void
	not map_dirty implies map.count = count
	count > 0 implies root /= Void and then root.count = count

end -- class AVL_TREE
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- This file is free software, which comes along with SmartEiffel. This software is distributed in the hope that it will be
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- You can modify it as you want, provided this footer is kept unaltered, and a notification of the changes is added.
-- You are allowed to redistribute it and sell it, alone or as a part of another product.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
