-- See the Copyright notice at the end of this file.
--
deferred class SET[E]
	--
	-- Definition of a mathematical set of objects. All common
	-- operations on mathematical sets are available.
	--
	-- Well knowned implementations, see HASHED_SET and AVL_SET.
	-- 

inherit
	ANY
		redefine is_equal, copy
		end

insert
	SAFE_EQUAL[E]
		redefine is_equal, copy
		end

feature {ANY} -- Counting:
	count: INTEGER
			-- Cardinality of the set (i.e. actual `count' of stored elements).

	is_empty: BOOLEAN is
			-- Is the set empty?
		do
			Result := count = 0
		ensure
			Result = (count = 0)
		end

feature {ANY} -- Adding and removing:
	add (e: like item) is
			-- Add a new item to the set: the mathematical definition of
			-- adding in a set is followed.
		require
			e /= Void
		deferred
		ensure
			added: has(e)
			not_in_then_added: not old has(e) implies count = old count + 1
			in_then_not_added: old has(e) implies count = old count
		end

	remove (e: like item) is
			-- Remove item `e' from the set: the mathematical definition of
			-- removing from a set is followed.
		require
			e /= Void
		deferred
		ensure
			removed: not has(e)
			not_in_not_removed: not old has(e) implies count = old count
			in_then_removed: old has(e) implies count = old count - 1
		end

	frozen clear is
		obsolete "Now use `clear_count' or `clear_count_and_capacity' (july 17th 2004)."
		do
			clear_count
		end

	clear_count is
			-- Empty the current set (`is_empty' is True after that call). If possible, the actual implementation 
			-- is supposed to keep its internal storage area in order to refill `Current' in an efficient way.
			-- See also `clear_count_and_capacity' to select the most appropriate.
		deferred
		ensure
			is_empty: count = 0
		end

	clear_count_and_capacity is
			-- Empty the current set (`is_empty' is True after that call). If possible, the actual implementation 
			-- is supposed to release its internal storage area for this memory to be used by other objects.
			-- See also `clear_count' to select the most appropriate.
		deferred
		ensure
			is_empty: count = 0
		end

feature {ANY} -- Looking and searching:
	has (e: like item): BOOLEAN is
			-- Is element `e' in the set?
		require
			e /= Void
		deferred
		ensure
			Result implies not is_empty
		end

	fast_has (e: like item): BOOLEAN is
			-- Is element `e' actually stored in the set?
		require
			e /= Void
		deferred
		ensure
			Result implies (e = reference_at(e))
		end

	reference_at (e: like item): like item is
			-- Non Void when `e' is in the set. In such a situation, `Result' is the object which is actually 
			-- stored in the `Current' set (see ensure assertion).
		require
			e /= Void
			not e.is_expanded_type
		deferred
		ensure
			has(e) implies Result.is_equal(e)
		end

feature {ANY} -- To provide iterating facilities:
	lower: INTEGER is 1

	upper: INTEGER is
		do
			Result := count
		ensure
			Result = count
		end

	valid_index (index: INTEGER): BOOLEAN is
		do
			Result := 1 <= index and then index <= count
		ensure
			Result = index.in_range(lower, upper)
		end

	item (index: INTEGER): E is
			-- Return the item indexed by `index'.
		require
			valid_index(index)
		deferred
		ensure
			has(Result)
		end

	get_new_iterator: ITERATOR[E] is
		do
			create {ITERATOR_ON_SET[E]} Result.make(Current)
		end

feature {ANY} -- Mathematical operations:
	union (other: like Current) is
			-- Make the union of the `Current' set with `other'.
		require
			other /= Void
		local
			i: INTEGER; e: like item
		do
			from
				i := 1
			until
				i > other.count
			loop
				e := other.item(i)
				if not has(e) then
					add(e)
				end
				i := i + 1
			end
		ensure
			count <= old count + other.count
		end

	infix "+" (other: like Current): like Current is
			-- Return the union of the `Current' set with `other'.
		require
			other /= Void
		do
			Result := twin
			Result.union(other)
		ensure
			Result.count <= count + other.count
			Current.is_subset_of(Result) and then other.is_subset_of(Result)
		end

	intersection (other: like Current) is
			-- Make the intersection of the `Current' set with `other'.
		require
			other /= Void
		local
			i: INTEGER; e: like item
		do
			from
				i := upper
			until
				i < lower
			loop
				e := item(i)
				if not other.has(e) then
					remove(e)
				end
				i := i - 1
			end
		ensure
			count <= other.count.min(old count)
		end

	infix "^" (other: like Current): like Current is
			-- Return the intersection of the `Current' set with `other'.
		require
			other /= Void
		do
			Result := twin
			Result.intersection(other)
		ensure
			Result.count <= other.count.min(count)
			Result.is_subset_of(Current) and then Result.is_subset_of(other)
		end

	minus (other: like Current) is
			-- Make the set `Current' - `other'.
		require
			other /= Void
		local
			i: INTEGER
		do
			if other = Current then
				clear_count
			else
				from
					i := 1
				until
					i > other.count
				loop
					remove(other.item(i))
					i := i + 1
				end
			end
		ensure
			count <= old count
		end

	infix "-" (other: like Current): like Current is
			-- Return  the set `Current' - `other'.
		require
			other /= Void
		do
			Result := twin
			Result.minus(other)
		ensure
			Result.count <= count
			Result.is_subset_of(Current)
		end

feature {ANY} -- Comparison:
	is_subset_of (other: like Current): BOOLEAN is
			-- Is the `Current' set a subset of `other'?
		require
			other /= Void
		local
			i: INTEGER
		do
			if Current = other then
				Result := True
			elseif count <= other.count then
				from
					Result := True
					i := 1
				until
					not Result or else i > count
				loop
					Result := other.has(item(i))
					i := i + 1
				end
			end
		ensure
			Result implies count <= other.count
		end

	is_disjoint_from (other: like Current): BOOLEAN is
			-- Is the `Current' set disjoint from `other' ?
		require
			other /= Void
		local
			i: INTEGER
		do
			if Current /= other then
				Result := True
				i := 1
				if count <= other.count then
					from
					until
						not Result or else i > count
					loop
						Result := not other.has(item(i))
						i := i + 1
					end
				else
					from
					until
						not Result or else i > other.count
					loop
						Result := not has(other.item(i))
						i := i + 1
					end
				end
			end
		ensure
			Result = (Current ^ other).is_empty
		end

	is_equal (other: like Current): BOOLEAN is
			-- Is the `Current' set equal to `other'?
		local
			i: INTEGER
		do
			if Current = other then
				Result := True
			elseif count = other.count then
				from
					Result := True
					i := 1
				until
					not Result or else i > count
				loop
					Result := other.has(item(i))
					i := i + 1
				end
			end
		ensure then
			double_inclusion: Result = (is_subset_of(other) and other.is_subset_of(Current))
		end

feature {ANY}
	copy (other: like Current) is
			-- Copy 'other' into the current set
		local
			i: INTEGER
		do
			-- Note: this naive implementation is OK because node 
			-- recycling is implemented.
			from
				make
				i := 1
			until
				i > other.count
			loop
				add(other.item(i))
				i := i + 1
			end
		end

	from_collection (model: COLLECTION[like item]) is
			-- Add all items of `model'.
		require
			model /= Void
		local
			i, up: INTEGER
		do
			from
				make
				up := model.upper
				i := model.lower
			until
				i > up
			loop
				add(model.item(i))
				i := i + 1
			end
		end

feature {ANY} -- Agents based features:
	do_all (action: ROUTINE[ANY, TUPLE[E]]) is
			-- Apply `action' to every item of `Current'.
		local
			i: INTEGER
		do
			from
				i := lower
			until
				i > upper
			loop
				action.call([item(i)])
				i := i + 1
			end
		end

	for_all (predicate: PREDICATE[ANY, TUPLE[E]]): BOOLEAN is
			-- Do all items satisfy `predicate'?
		local
			i: INTEGER
		do
			from
				Result := True
				i := lower
			until
				not Result or else i > upper
			loop
				Result := predicate.item([item(i)])
				i := i + 1
			end
		end

	exists (predicate: PREDICATE[ANY, TUPLE[E]]): BOOLEAN is
			-- Does at least one item satisfy `predicate'?
		local
			i: INTEGER
		do
			from
				i := lower
			until
				Result or else i > upper
			loop
				Result := predicate.item([item(i)])
				i := i + 1
			end
		end

feature {}
	make is
			-- Creation of an empty SET.
		deferred
		ensure
			is_empty
		end

feature {}
	-- Implement manifest generic creation.

	manifest_make (needed_capacity: INTEGER) is
			-- Manifest creation of a set.
		do
			make
		end
	
	manifest_put (index: INTEGER; element: like item) is
		do
			add(element)
		end	

	manifest_semicolon_check: BOOLEAN is False
	
end -- class SET
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- This file is free software, which comes along with SmartEiffel. This software is distributed in the hope that it will be
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- You can modify it as you want, provided this footer is kept unaltered, and a notification of the changes is added.
-- You are allowed to redistribute it and sell it, alone or as a part of another product.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
