-- See the Copyright notice at the end of this file.
--
class AVL_SET[E -> COMPARABLE]
	--
	-- Definition of a mathematical set of comparable objects. All common
	-- operations on mathematical sets are available.
	--

inherit
	SET[E]
	AVL_TREE[E]
		redefine
			root
		end

creation {ANY}
	make, from_collection, manifest_creation

feature {ANY}
	add (e: like item) is
			-- Add a new item to the set: the mathematical definition of
			-- adding in a set is followed.
		do
			root := do_insert(root, e)
		end

	clear_count, clear_count_and_capacity is
		do
			if not is_empty then
				clear_nodes(root)
				root := Void
				count := 0
				map_dirty := True
			end
		end

	reference_at (e: like item): like item is
		local
			n: AVL_SET_NODE[E]
		do
			if root /= Void then
				n := root.at(e)
				if n /= Void then
					Result := n.item
				end
			end
		end

feature {ANY} -- To provide iterating facilities:
	item (index: INTEGER): E is
			-- Return the item indexed by `index'.
		do
			if map_dirty then
				build_map
			end
			Result := map.item(index - 1).item
		end

feature {}
	root: AVL_SET_NODE[E]

	do_insert (n: like a_new_node; e: like item): like a_new_node is
		do
			if n = Void then
				Result := new_node
				Result.make(e)
				count := count + 1
				map_dirty := True
				rebalance := True
			elseif e < n.item then
				n.set_left(do_insert(n.left, e))
				if rebalance then
					Result := left_grown(n)
				else
					Result := n
				end
			elseif e > n.item then
				n.set_right(do_insert(n.right, e))
				if rebalance then
					Result := right_grown(n)
				else
					Result := n
				end
			else
				Result := n
				rebalance := False
			end
		ensure
			Result /= Void
			Result.balance = node_height(Result.right) - node_height(Result.left)
			rebalance = (node_height(Result) > old node_height(n))
		end

	exchange_and_discard (n1, n2: like a_new_node) is
		do
			map_dirty := True
			n1.set_item(n2.item)
			rebalance := True
			count := count - 1
			discard_node(n2)
		end

	new_node: like a_new_node is
		do
			Result := lost_nodes.item
			if Result = Void then
				Result := a_new_node
			else
				lost_nodes.set_item(Result.left)
			end
		end

	discard_node (n: like a_new_node) is
		local
			i: like item
		do
			n.make(i)
			n.set_left(lost_nodes.item)
			lost_nodes.set_item(n)
		end

	a_new_node: AVL_SET_NODE[like item] is
		do
			create Result
		end

	lost_nodes: WEAK_REFERENCE[like a_new_node]

	common_lost_nodes: DICTIONARY[WEAK_REFERENCE[ANY_AVL_SET_NODE], STRING] is
		once
			create {HASHED_DICTIONARY[WEAK_REFERENCE[ANY_AVL_SET_NODE], STRING]} Result.make
		end

feature {}
	make is
		do
			if lost_nodes /= Void then
				clear_count
			else
				create map.make(0)
				lost_nodes ::= common_lost_nodes.reference_at(generating_type)
				if lost_nodes = Void then
					create lost_nodes.set_item(Void)
					common_lost_nodes.add(lost_nodes, generating_type)
				end
			end
		end

invariant
	lost_nodes /= Void
	lost_nodes = common_lost_nodes.at(generating_type)

end -- class AVL_SET
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- This file is free software, which comes along with SmartEiffel. This software is distributed in the hope that it will be
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- You can modify it as you want, provided this footer is kept unaltered, and a notification of the changes is added.
-- You are allowed to redistribute it and sell it, alone or as a part of another product.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
