/*
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- This file is free software, which comes along with SmartEiffel. This software is distributed in the hope that it will be
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- You can modify it as you want, provided this footer is kept unaltered, and a notification of the changes is added.
-- You are allowed to redistribute it and sell it, alone or as a part of another product.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
*/
static char net_error[1024];
char* net_last_error = NULL;
int net_last_error_number;


#if defined __USE_POSIX || defined __unix__ || defined _POSIX_C_SOURCE

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>

static clear_error() {
  net_last_error = NULL;
  net_last_error_number = 0;
}

static set_error() {
  if (errno == 0) {
    clear_error();
  }
  else {
    net_last_error = net_error;
    net_last_error_number = errno;
    sprintf(net_last_error, "%d: %s", errno, strerror(errno));
    fprintf(SE_ERR, "***** %s\n", net_last_error);
  }
}

static set_host_error() {
  char* s;
  if (h_errno == 0) {
    clear_error();
  }
  else {
    net_last_error = net_error;
    net_last_error_number = h_errno;
    switch(h_errno) {
    case HOST_NOT_FOUND:
      s = "Unknown host";
      break;
    case NO_ADDRESS:
      s = "Unknown IP";
      break;
    case NO_RECOVERY:
      s = "Non-recoverable name server error";
      break;
    case TRY_AGAIN:
      s = "Server error, try again later";
      break;
    default:
      s = "Unknown error";
    }
    sprintf(net_last_error, "%d: %s", h_errno, s);
    fprintf(SE_ERR, "***** %s\n", net_last_error);
  }
}

int net_hostname(void* a_hostname) {
  struct hostent* host;
  int result = 0;

  clear_error();
  host = gethostbyname((char*)a_hostname);
  if (host != NULL) {
    result = ntohl(((struct in_addr*)(host->h_addr))->s_addr);
  }
  else {
    set_host_error();
  }

  return result;
}

static int control_socket(int socket) {
  int result = socket;
  int flags;
  clear_error();
  /*
  if (socket != -1) {
    flags = fcntl (result, F_GETFL) ;
    flags |= O_NONBLOCK ;
    if (fcntl (result, F_SETFL, flags) == -1) {
      result = -1;
      set_error();
    }
  }
  */

  return result;
}

static int address_socket(int socket, int a, int b, int c, int d, int port) {
  static struct sockaddr_in tmp_addr;
  int result = socket;
  clear_error();
  tmp_addr.sin_family = AF_INET;
  tmp_addr.sin_port = htons(port);
  tmp_addr.sin_addr.s_addr = htonl((((((a<<8) + b)<<8) + c)<<8) + d) ;
  if (connect(result, (struct sockaddr *)&tmp_addr, sizeof(struct sockaddr_in)) < 0 && errno != EINPROGRESS) {
    result = -1;
    set_error();
  }
  set_error();
  return result;
}

int net_tcp(int a, int b, int c, int d, int port) {
  int result = 0;
  result = net_tcp_socket();
  if (result >= 0) {
    result = address_socket(result, a, b, c, d, port);
  }
  return result;
}

int net_udp(int a, int b, int c, int d, int port) {
  int result = 0;
  result = net_udp_socket();
  if (result >= 0) {
    result = address_socket(result, a, b, c, d, port);
  }
  return result;
}

int net_local(int port) {
  int result = 0;
  static struct sockaddr_in tmp_addr;
  clear_error();
  result = net_local_socket();
  tmp_addr.sin_family = AF_LOCAL;
  tmp_addr.sin_port = htons(port);
  if (connect(result, (struct sockaddr *)&tmp_addr, sizeof(struct sockaddr_in)) < 0 && errno != EINPROGRESS) {
    result = -1;
    set_error();
  }
  return result;
}

void net_disconnect(int fd) {
  int ret = close(fd);
  clear_error();
  if (ret == -1) {
    set_error();
  }
}

int net_tcp_socket() {
  int result = socket(PF_INET, SOCK_STREAM, 0);
  clear_error();
  if (result == -1) {
    set_error();
  }
  else {
    result = control_socket(result);
  }
  return result;
}

int net_udp_socket() {
  int result = socket(PF_INET, SOCK_DGRAM, 0);
  clear_error();
  if (result == -1) {
    set_error();
  }
  else {
    result = control_socket(result);
  }
  return result;
}

int net_local_socket() {
  int result = socket(PF_LOCAL, SOCK_STREAM, 0);
  clear_error();
  if (result == -1) {
    set_error();
  }
  else {
    result = control_socket(result);
  }
  return result;
}

static int bind_server(int socket, int port, int sockfamily) {
  int result = socket;
  struct sockaddr_in local;
  socklen_t size = sizeof(struct sockaddr_in);
  int ret;
  local.sin_family = sockfamily;
  local.sin_port = htons(port);
  local.sin_addr.s_addr = htonl(INADDR_ANY);
  ret = bind(result, (struct sockaddr*)&local, size);
  if (ret != 0) {
    set_error();
    result = -1;
  }
  else {
    ret = listen(result, 0);
    if (ret != 0) {
      set_error();
      result = -1;
    }
  }
  return result;
}

int net_tcp_server(int port) {
  return bind_server(net_tcp_socket(), port, AF_INET);
}

int net_udp_server(int port) {
  return bind_server(net_udp_socket(), port, AF_INET);
}

int net_local_server(int port) {
  return bind_server(net_local_socket(), port, AF_LOCAL);
}

void net_select(int count, int* afd, float timeout) {
  static fd_set tmp_set;
  int r;
  int maxfd = 0;
  int i, fd;
  struct timeval tv0 ;
  struct timeval *tv = NULL;

  clear_error();

  FD_ZERO(&tmp_set);
  for (i = 0; i < count; i++) {
    fd = afd[i];
    FD_SET(fd, &tmp_set);
    if (fd > maxfd) {
      maxfd = fd;
    }
  }

  if (timeout >= 0) {
    tv0.tv_sec = (int)timeout;
    tv0.tv_usec = (int)(1000000*(timeout-tv0.tv_sec)) ;
    tv = &tv0 ;
  }

  r = select(maxfd + 1, &tmp_set, NULL, NULL, tv);

  if (r > 0) {
    for (i = 0; i < count; i++) {
      fd = afd[i];
      if (!FD_ISSET(fd, &tmp_set)) {
        afd[i] = -1;
      }
    }
  }
  else {
    set_error();
  }
}

int net_select_one(int fd, float timeout) {
  int result = fd;
  int afd[1];
  afd[0] = fd;
  net_select(1, afd, timeout);
  result = afd[0];
  return result;
}

int net_read(int fd, int count, char* buffer) {
  int result = 0;
  clear_error();
  result = read(fd, buffer, count);
  if (result < 0) {
    set_error();
  }
  return result;
}

int net_write(int fd, int count, char* buffer) {
  int result = 0;
  int charcount;
  clear_error();
  while (result >= 0 && count > 0) {
    charcount = write(fd, buffer + result, count);
    if (charcount < 0) {
      result = -1;
      set_error();
    }
    else {
      count -= charcount;
      result += charcount;
    }
  }
  return result;
}

void net_accept(int server_fd, int* out_values) {
  static struct sockaddr_in tmp_addr;
  int ip, a, b, c, d, port, fd;
  socklen_t size = sizeof(struct sockaddr_in);
  clear_error();
  fd = accept(server_fd, (struct sockaddr*)&tmp_addr, &size);
  if (fd >= 0) {
    port = ntohs(tmp_addr.sin_port) ;
    ip = ntohl(tmp_addr.sin_addr.s_addr);
    a = (ip >> 24) & 0xff;
    b = (ip >> 16) & 0xff;
    c = (ip >>  8) & 0xff;
    d =  ip        & 0xff;
  }
  else {
    set_error();
  }
  out_values[0]=a;
  out_values[1]=b;
  out_values[2]=c;
  out_values[3]=d;
  out_values[4]=port;
  out_values[5]=fd;
}

#elif defined WIN32

int net_hostname(void* a_hostname) {
}

static int control_socket(int socket) {
}

static int address_socket(int socket, int a, int b, int c, int d, int port) {
}

int net_tcp(int a, int b, int c, int d, int port) {
}

int net_udp(int a, int b, int c, int d, int port) {
}

int net_local(int port) {
}

void net_disconnect(int fd) {
}

int net_tcp_socket() {
}

int net_udp_socket() {
}

int net_local_socket() {
}

int net_tcp_server(int port) {
}

int net_udp_server(int port) {
}

int net_local_server(int port) {
}

void net_select(int count, int* afd, float timeout) {
}

int net_select_one(int fd, float timeout) {
}

int net_read(int fd, int count, char* buffer) {
}

int net_write(int fd, int count, char* buffer) {
}

void net_accept(int server_fd, int* out_values) {
}

#else
#endif
