-- This file is part of SmartEiffel The GNU Eiffel Compiler Tools and Libraries.
-- See the Copyright notice at the end of this file.
--
class AGENT_CREATION
	--
	-- An Eiffel agent is a delayed call. Such an Eiffel object include a feature call with closed or open
	-- arguments (or closed or open target as well). An AGENT_CREATION object is the notation to create the
	-- agent (i.e. the feature is not launched yet, see also AGENT_EXPRESSION).
	--
	-- Examples of AGENT_CREATION:
	--          agent my_array.add_last(?)         -- The target is closed and the argument is open.
	--          agent {ARRAY[INTEGER]}.add_last(1) -- The target is open and the argument is closed.
	--          agent {ARRAY[INTEGER]}.add_last(?) -- Both the target and the argument are open.
	--          agent {ARRAY[INTEGER]}.add_last    -- A short-hand for the previous notation.
	--          agent my_array.put(?, 3)           -- Only the second argument is open.
	--          agent my_array.put(3, ?)           -- Only the first argument is open.
	--          agent {ARRAY[INTEGER]}.put         -- Target and all arguments are open.
	--

inherit
	EXPRESSION
		redefine is_equal
		end
	HASHABLE
		redefine is_equal
		end

creation {EIFFEL_PARSER}
	make

feature {ANY}
	start_position: POSITION
			-- Of the first letter of the mandatory "agent" keyword (obviously provided by the 
			-- `eiffel_parser').

	delayed_call: CALL
			-- Syntactically, the call to be delayed (it can become a procedure call as well).
	
	delayed_proc_call: PROC_CALL
			-- Non Void when `delayed_call' is finally a procedure call instead.
	
feature {ANY}
	is_current, is_implicit_current: BOOLEAN is False

	extra_bracket_flag: BOOLEAN is False

	hash_code: INTEGER is
		do
			Result := delayed_call.feature_name.hash_code
		end

	is_equal (other: like Current): BOOLEAN is
		do
			if Current = other then
				Result := True
			elseif delayed_proc_call /= Void then
				Result := delayed_proc_call = other.delayed_proc_call
			else
				Result := delayed_call = other.delayed_call
			end
		end

	declaration_type: TYPE is
		do
			Result := resolve_in(start_position.class_text.declaration_type_of_like_current)
		end

	specialize_in (type: TYPE): like Current is
		do
			if delayed_proc_call /= Void then
				-- Because of genericity, it is possible:
				Result := current_or_twin_init(delayed_call, delayed_proc_call.specialize_in(type))
			else
				Result := current_or_twin_init(delayed_call.specialize_in(type), delayed_proc_call)
			end
		end

	specialize_thru (parent_type: TYPE; inherit_item: INHERIT_TEXT_ITEM; new_type: TYPE): like Current is
		local
			dc: like delayed_call; dpc: like delayed_proc_call
		do
			if delayed_proc_call /= Void then
				dpc := delayed_proc_call.specialize_thru(parent_type, inherit_item, new_type)
				Result := current_or_twin_init(delayed_call, dpc)
			else
				dc := delayed_call.specialize_thru(parent_type, inherit_item, new_type)
				Result := current_or_twin_init(dc, Void)
			end
		end

	specialize_2 (type: TYPE): like Current is
		local
			dc: like delayed_call; dpc: like delayed_proc_call; target: EXPRESSION;
			arguments: EFFECTIVE_ARG_LIST; fs: FEATURE_STAMP; af: ANONYMOUS_FEATURE; target_type: TYPE
		do
			if delayed_proc_call /= Void then
				target := delayed_proc_call.target.specialize_2(type)
			else
				target := delayed_call.target.specialize_2(type)
			end
			fs := target.declaration_type.search(delayed_call.feature_name)
			if fs = Void then
				smart_eiffel.unknown_feature_fatal_error(target, target.declaration_type, delayed_call.feature_name)
			end
			target_type := target.resolve_in(type)
			fs := fs.resolve_dynamic_binding_for(target_type)
			af := fs.anonymous_feature(target_type)
			if af.arguments = Void then
				smart_eiffel.argument_count_check(start_position, af, delayed_call.arguments)
			elseif delayed_call.arguments = Void then
				arguments := af.arguments.omitted_open_arguments(delayed_call.feature_name.start_position)
				if arguments.count = 1 then
					create {CALL_1_C} delayed_call.make(target, delayed_call.feature_name, arguments)
				   delayed_call.set_feature_stamp(fs)
				else
					create {CALL_N} delayed_call.make(target, delayed_call.feature_name, arguments)
				   delayed_call.set_feature_stamp(fs)
				end
			else
				arguments := delayed_call.arguments
				smart_eiffel.argument_count_check(start_position, af, arguments)
			end
			check
				af.arguments /= Void implies af.arguments.count = delayed_call.arguments.count
			end
			agent_creation_helper.resolve_open_operands(type, target_type, af, target, arguments)
			if arguments /= Void then
				arguments := arguments.specialize_2(type, af, target_type, delayed_call.target.is_current)
			end
			if af.result_type = Void then
				-- It is actually a procedure call:
				dc := delayed_call
				if delayed_proc_call /= Void and then
					delayed_proc_call.target = target and then
					delayed_proc_call.arguments = arguments and then
					delayed_proc_call.feature_stamp = fs
				 then
					-- Yes we can recycle:
					dpc := delayed_proc_call
				elseif arguments = Void then
					create {PROC_CALL_0} dpc.make(target, delayed_call.feature_name)
					dpc.set_feature_stamp(fs)
				elseif arguments.count = 1 then
					create {PROC_CALL_1} dpc.make(target, delayed_call.feature_name, arguments)
					dpc.set_feature_stamp(fs)
				else
					create {PROC_CALL_N} dpc.make(target, delayed_call.feature_name, arguments)
					dpc.set_feature_stamp(fs)
				end
				-- To get all exportation checks done too:
				dpc ::= dpc.specialize_2(type)
			else
				-- To get all exportation checks done too:
				if delayed_call.target = target and then
					delayed_call.arguments = arguments and then
					delayed_call.feature_stamp = fs
				 then
					dc := delayed_call
				else
					dc := delayed_call.twin
					dc.set_target(target)
					if arguments /= Void then
						dc.set_arguments(arguments)
					end
					dc.set_feature_stamp(fs)
				end
				dc ::= dc.specialize_2(type)
			end
			Result := current_or_twin_init(dc, dpc)
		end

	has_been_specialized: BOOLEAN is
		do
			if delayed_proc_call /= Void then
				Result := delayed_proc_call.has_been_specialized
			else
				Result := delayed_call.has_been_specialized
			end
		end

	resolve_in (type: TYPE): TYPE is
		local
			target: EXPRESSION; a: EFFECTIVE_ARG_LIST; actual_base, open_tuple, rt: TYPE_MARK;
			target_type: TYPE; open_type_list: ARRAY[TYPE_MARK]; expression: EXPRESSION
			af: ANONYMOUS_FEATURE; i: INTEGER; fs: FEATURE_STAMP
		do
			if delayed_proc_call /= Void then
				target := delayed_proc_call.target
				a := delayed_proc_call.arguments
				fs := delayed_proc_call.feature_stamp
			else
				target := delayed_call.target
				a := delayed_call.arguments
				fs := delayed_call.feature_stamp
			end
			update_argument_list(type, target, a)
			from
				i := argument_list.lower
			until
				i > argument_list.upper
			loop
				expression := argument_list.item(i)
				if expression.is_open_target_or_open_operand then
					if open_type_list = Void then
						create open_type_list.with_capacity(4, 1)	-- *** Still leaky. ***
					end
					open_type_list.add_last(expression.resolve_in(type).canonical_type_mark)
				end
				i := i + 1
			end
			if open_type_list /= Void then
				create {NON_EMPTY_TUPLE_TYPE_MARK} open_tuple.make(start_position, open_type_list)
			else
				create {EMPTY_TUPLE_TYPE_MARK} open_tuple.make(start_position)
			end
			check
				open_tuple.is_static
			end
			target_type := target.resolve_in(type)
			af := fs.anonymous_feature(target_type)
			rt := af.result_type
			actual_base := target_type.canonical_type_mark
			if rt = Void then
				create {AGENT_TYPE_MARK} rt.procedure(start_position, actual_base, open_tuple)
			else
				rt := rt.to_static(target_type)
				if rt.is_boolean then
					create {AGENT_TYPE_MARK} rt.predicate(start_position, actual_base, open_tuple)
				else
					create {AGENT_TYPE_MARK} rt.function(start_position, actual_base, open_tuple, rt)
				end
			end
			check
				rt.is_static
			end
			Result := rt.type
		end

	collect (type: TYPE): TYPE is
		local
			unused: TYPE
		do
			if delayed_proc_call /= Void then
				delayed_proc_call.collect(type)
			else
				unused := delayed_call.collect(type)
			end
			Result := resolve_in(type)
			smart_eiffel.collect_create(Result)
			agent_pool.agent_creation_collect(type, Current, Result)
		end

	adapt_for (type: TYPE): like Current is
		local
			dc: like delayed_call; dpc: like delayed_proc_call
		do
			if delayed_proc_call /= Void then
				dpc := delayed_proc_call.adapt_for(type)
				Result := current_or_twin_init(delayed_call, dpc)
			else
				dc := delayed_call.adapt_for(type)
				Result := current_or_twin_init(dc, Void)
			end
		end

	non_void_no_dispatch_type (type: TYPE): TYPE is
		do
			Result := resolve_in(type)
		end

	simplify (type: TYPE): like Current is
		local
			target: EXPRESSION; arguments: EFFECTIVE_ARG_LIST; dc: like delayed_call; dpc: like delayed_proc_call
		do
			if delayed_proc_call /= Void then
				target := delayed_proc_call.target.simplify(type)
				arguments := delayed_proc_call.arguments
				if arguments /= Void then
					arguments := arguments.simplify(type)
				end
				if target = delayed_proc_call.target and then arguments = delayed_proc_call.arguments then
					Result := Current
				else
					smart_eiffel.magic_count_increment
					dpc := delayed_proc_call.twin
					dpc.set_target(target)
					dpc.set_arguments(arguments)
					Result := twin
					Result.set_delayed_proc_call(dpc)
				end
			else
				target := delayed_call.target.simplify(type)
				arguments := delayed_call.arguments
				if arguments /= Void then
					arguments := arguments.simplify(type)
				end
				if target = delayed_call.target and then arguments = delayed_call.arguments then
					Result := Current
				else
					smart_eiffel.magic_count_increment
					dc := delayed_call.twin
					dc.set_target(target)
					dc.set_arguments(arguments)
					Result := twin
					Result.set_delayed_call(dc)
				end
			end
		end

	accept (visitor: AGENT_CREATION_VISITOR) is
		do
			visitor.visit_agent_creation(Current)
		end

	is_void: BOOLEAN is False

	is_writable: BOOLEAN is False

	is_manifest_string: BOOLEAN is False

	is_static: BOOLEAN is False

	is_result: BOOLEAN is False

	side_effect_free (type: TYPE): BOOLEAN is
		do
			-- Memory allocation.
		end

	safety_check (type: TYPE) is
		do
			if delayed_proc_call /= Void then
				delayed_proc_call.safety_check(type)
			else
				delayed_call.safety_check(type)
			end
		end

	compile_to_c (type: TYPE) is
		local
			buffer: STRING; i: INTEGER; boost, need_comma: BOOLEAN; target, expression: EXPRESSION
			arguments: EFFECTIVE_ARG_LIST
		do
			buffer:= once "......................"
			buffer.clear_count
			mold_id_in(type, buffer)
			buffer.extend('(')
			boost := ace.boost
			if not boost then
				buffer.append(once "&ds")
				need_comma := True
			end	
			if delayed_proc_call /= Void then
				target := delayed_proc_call.target
				arguments := delayed_proc_call.arguments
			else
				target := delayed_call.target
				arguments := delayed_call.arguments
			end
			update_argument_list(type, target, arguments)
			cpp.put_string(buffer)
			from
				i := argument_list.lower
			until
				i > argument_list.upper
			loop
				expression := argument_list.item(i)
				if not expression.is_open_target_or_open_operand then
					if need_comma then
						cpp.put_character(',')
					end
					expression.compile_to_c(type)
					need_comma := True
				end
				i := i + 1
			end
			cpp.put_character(')')
		end

	mapping_c_target (type: TYPE; formal_type: TYPE_MARK) is
		do
			compile_to_c(type)
		end

	mapping_c_arg (type: TYPE; formal_arg_type: TYPE_MARK) is
		do
			compile_to_c(type)
		end

	use_current (type: TYPE): BOOLEAN is
		do
			if delayed_proc_call /= Void then
				Result := delayed_proc_call.use_current(type)
			else
				Result := delayed_call.use_current(type)
			end
		end

	bracketed_pretty, pretty (indent_level: INTEGER) is
		do
			pretty_printer.keyword(once "agent")
			delayed_call.pretty(indent_level)
		end

	pretty_target (indent_level: INTEGER) is
		do
			pretty_printer.put_character('(')
			pretty(indent_level)
			pretty_printer.put_character(')')
			pretty_printer.put_character('.')
		end

	short (type: TYPE) is
		do
			short_printer.hook_or(as_agent, as_agent)
			short_printer.put_character(' ')
			delayed_call.short(type)
		end

	short_target (type: TYPE) is
		do
			bracketed_short(type)
			short_printer.put_dot
		end

	precedence: INTEGER is
		do
			Result := atomic_precedence
		end

	jvm_assign_creation, jvm_assign (type: TYPE) is
		do
			check
				False
			end
		end

	compile_target_to_jvm, compile_to_jvm (type: TYPE) is
		do
			not_yet_implemented
		end

	jvm_branch_if_false, jvm_branch_if_true (type: TYPE): INTEGER is
		do
			not_yet_implemented
		end

feature {AGENT_POOL}
	mold_id_in (type: TYPE; buffer: STRING) is
			-- Identify the corresponding AGENT_CREATION function .
		do
			buffer.extend('T')
			type.live_type.id.append_in(buffer)
			buffer.extend('C')
			start_position.class_text.id.append_in(buffer)
			buffer.extend('l')
			start_position.line.append_in(buffer)
			buffer.extend('c')
			start_position.column.append_in(buffer)
		end
	
	c_define_function (type: TYPE; mold_id: STRING; integer_mold_id: INTEGER) is
		require
			cpp.on_c
		local
			boost: BOOLEAN; target_type, agent_type, agent_result: TYPE; tm: TYPE_MARK;
			i: INTEGER; buffer: STRING; rf: RUN_FEATURE; expression: EXPRESSION
			user_expanded_flag: BOOLEAN; fs: FEATURE_STAMP
		do
			boost := ace.boost
			cpp.swap_on_h
			agent_type := c_define_type(type, boost, mold_id, integer_mold_id)
			cpp.swap_on_c
			agent_result := agent_type.agent_result
			-- The local execution function:
			buffer := once "............................ unique buffer .........................."
			buffer.copy(once "%N/*agent launcher*/static ")
			if agent_result /= Void then
				agent_result.canonical_type_mark.c_type_for_result_in(buffer)
			else
				buffer.append(fz_void)
			end
			buffer.append(once " _")
			buffer.append(mold_id)
			buffer.extend('(')
			if not boost then
				buffer.append(once "se_dump_stack*caller,")
			end
			if ace.profile then
				buffer.append(once "se_local_profile_t*parent_profile,")
			end
			buffer.append(once "se_")
			buffer.append(mold_id)
			buffer.append(once "*u")
			from
				i := argument_list.lower
			until
				i > argument_list.upper
			loop
				expression := argument_list.item(i)
				if expression.is_open_target_or_open_operand then
					buffer.extend(',')
					expression.resolve_in(type).canonical_type_mark.c_type_for_result_in(buffer)
					buffer.extend(' ')
					agent_creation_helper.compile_to_c_in(buffer, type, expression)					
				end
				i := i + 1
			end
			buffer.append(once "){%N")
			from
				i := argument_list.lower
			until
				i > argument_list.upper
			loop
				expression := argument_list.item(i)
				if not expression.is_open_target_or_open_operand then
					agent_creation_helper.same_c_type_in(buffer, type, expression)
					if expression.is_current and then type.is_user_expanded then
						buffer.extend('*')
						user_expanded_flag := True
					else
						buffer.extend(' ')
						user_expanded_flag := False
					end
					agent_creation_helper.compile_to_c_in(buffer, type, expression)
					buffer.extend('=')
					if user_expanded_flag then
						buffer.extend('&')
					end					
					buffer.append(once "(u->")
					agent_creation_helper.compile_to_c_in(buffer, type, expression)
					buffer.append(once ");%N")
				end
				i := i + 1
			end
			cpp.put_string(buffer)
			if ace.profile then
				smart_eiffel.local_profile
			end
			if not boost then
				cpp.put_string(once "se_dump_stack ds;%Nds=(*caller);%N")
			end
			-- Calling the actual one:
			if ace.profile then
				smart_eiffel.start_profile_agent_creation(Current)
			end
			if agent_result /= Void then
				cpp.put_string(once "u->R=")
			end
			if delayed_proc_call /= Void then
				target_type := delayed_proc_call.target.resolve_in(type)
				fs := delayed_proc_call.feature_stamp
				rf := fs.run_feature_for(target_type)
			else
				target_type := delayed_call.target.resolve_in(type)
				fs := delayed_call.feature_stamp
				rf := fs.run_feature_for(target_type)
			end
			if delayed_proc_call /= Void then
				cpp.put_cpc(type, fs, rf, delayed_proc_call.target, delayed_proc_call.arguments)
			else
				cpp.put_cpc(type, fs, rf, delayed_call.target, delayed_call.arguments)
			end
			if agent_result /= Void then
				cpp.put_string(once ";%N")
			end
			if ace.profile then
				smart_eiffel.stop_profile
			end
			if agent_result /= Void then
				cpp.put_string(once "return u->R;%N") -- ***** RESULT R ??? *****
			end
			cpp.put_string(once "}%N")
			-- The definition function:
			buffer.copy(once "/*agent creation*/T0*")
			buffer.append(mold_id)
			buffer.extend('(')
			if not boost then
				buffer.append(once "se_dump_stack*caller")
			end
			from
				i := argument_list.lower
			until
				i > argument_list.upper
			loop
				expression := argument_list.item(i)
				if not expression.is_open_target_or_open_operand then
					if buffer.last /= '(' then
						buffer.extend(',')
					end
					agent_creation_helper.same_c_type_in(buffer, type, expression)
					buffer.extend(' ')
					agent_creation_helper.compile_to_c_in(buffer, type, expression)
				end
				i := i + 1
			end
			if buffer.last = '(' then
				buffer.append(once "void")
			end
			buffer.extend(')')
			cpp.put_string_on_h(buffer)
			cpp.put_string_on_h(once ";%N")
			buffer.append(once "{%Nse_")
			buffer.append(mold_id)
			buffer.append(once "*u=(void*)new_agent(")
			agent_type.live_type.id.append_in(buffer)
			buffer.append(once ");%Nu->creation_mold_id=")
			integer_mold_id.append_in(buffer)
			buffer.append(once ";%Nu->afp=_")
			buffer.append(mold_id)
			buffer.append(once ";%N")
			if not gc_handler.is_off then
				buffer.append(once "u->gc_mark_agent_mold=gc_mark_")
				buffer.append(mold_id)
				buffer.append(once ";%N")
			end
			from
				i := argument_list.lower
			until
				i > argument_list.upper
			loop
				expression := argument_list.item(i)
				if not expression.is_open_target_or_open_operand then
					buffer.append(once "u->")
					agent_creation_helper.compile_to_c_in(buffer, type, expression)
					buffer.extend('=')
					agent_creation_helper.compile_to_c_in(buffer, type, expression)
					buffer.append(once ";%N")
				end
				i := i + 1
			end
			buffer.append(once "return((T0*)u);%N}%N")
			cpp.put_string(buffer)
			-- The gc_mark_MOLD_ID function:
			if not gc_handler.is_off then
				buffer.copy(once "void gc_mark_")
				buffer.append(mold_id)
				buffer.append(once "(se_")
				buffer.append(mold_id)
				buffer.append(once "*u)")
				cpp.put_c_heading(buffer)
				buffer.copy(once "gc_agent*gcu=(gc_agent*)u;%N%
									  %if (gcu->header.flag==FSOH_UNMARKED){%N%
									  %gcu->header.flag=FSOH_MARKED;%N")
				from
					i := argument_list.lower
				until
					i > argument_list.upper
				loop
					expression := argument_list.item(i)
					if not expression.is_open_target_or_open_operand then
						tm := expression.resolve_in(type).canonical_type_mark
						if tm.is_reference then
							buffer.append(once "gc_mark(u->")
							agent_creation_helper.compile_to_c_in(buffer, type, expression)
							buffer.append(once ");%N")
						elseif tm.need_gc_mark_function then
							tm.gc_mark_in(buffer)
							buffer.append(once "(&(u->")
							agent_creation_helper.compile_to_c_in(buffer, type, expression)
							buffer.append(once "));%N")
						end
					end
					i := i + 1
				end
				buffer.append(once "}%N}%N")
				cpp.put_string(buffer)
			end
		ensure
			cpp.on_c
		end

feature {AGENT_CREATION}
	init (dc: like delayed_call; dpc: like delayed_proc_call) is
		require
			dc /= Void
		do
			delayed_call := dc
			delayed_proc_call := dpc
		ensure
			delayed_call = dc
			delayed_proc_call = dpc
		end

	set_delayed_call (dc: like delayed_call) is
		require
			dc /= Void
		do
			delayed_call := dc
		ensure
			delayed_call = dc
		end
	
	set_delayed_proc_call (dpc: like delayed_proc_call) is
		require
			dpc /= Void
		do
			delayed_proc_call := dpc
		ensure
			delayed_proc_call = dpc
		end
	
feature {}
	make (sp: like start_position; dc: like delayed_call) is
		require
			not sp.is_unknown
			dc /= Void
		do
			start_position := sp
			delayed_call := dc
		ensure
			start_position = sp
			delayed_call = dc
		end
	
	current_or_twin_init (dc: like delayed_call; dpc: like delayed_proc_call): like Current is
		require
			dc /= Void
		do
			if dc = delayed_call and then dpc = delayed_proc_call then
				Result := Current
			else
				Result := twin
				Result.init(dc, dpc)
			end
		ensure
			Result.delayed_call = dc
			Result.delayed_proc_call = dpc
		end

	c_define_type (type: TYPE; boost: BOOLEAN; mold_id: STRING; integer_mold_id: INTEGER): TYPE is
			-- The Result is the `agent_type'. Also let the `argument_list' ready for use.
		require
			cpp.on_h
			boost = ace.boost
		local
			i: INTEGER; agent_type, t: TYPE; buffer: STRING; expression: EXPRESSION
		do
			-- The following call also update `argument_list', so it must be done first:
			agent_type := resolve_in(type)
			Result := agent_type
			
			buffer := once "............................ unique buffer .........................."
			buffer.copy(once "typedef struct _se_")
			buffer.append(mold_id)
			buffer.append(once " se_")
			buffer.append(mold_id)
			buffer.append(once ";%Nstruct _se_")
			buffer.append(mold_id)
			buffer.append(once "{Tid id;%Nint creation_mold_id;%N")
			t := agent_type.agent_result
			if t = Void then
				buffer.append(fz_void)
			else
				t.canonical_type_mark.c_type_for_result_in(buffer)
			end
			buffer.append(once "(*afp)(")
			if not boost then
				buffer.append(once "se_dump_stack*,")
			end
			if ace.profile then
				buffer.append(once "se_local_profile_t*,")
			end
			buffer.append(once "se_")
			buffer.append(mold_id)
			buffer.extend('*')
			from
				i := argument_list.lower
			until
				i > argument_list.upper
			loop
				expression := argument_list.item(i)
				if expression.is_open_target_or_open_operand then
					buffer.extend(',')
					t := expression.resolve_in(type)
					t.canonical_type_mark.c_type_for_result_in(buffer)
				end
				i := i + 1
			end
			buffer.append(once "[
			     );
			     void(*gc_mark_agent_mold)(se_
			     ]")
			buffer.append(mold_id)
			buffer.append(once "*);%N")
			from
				i := argument_list.lower
			until
				i > argument_list.upper
			loop
				expression := argument_list.item(i)
				if not expression.is_open_target_or_open_operand then
					agent_creation_helper.same_c_type_in(buffer, type, expression)
					buffer.extend(' ')
					agent_creation_helper.compile_to_c_in(buffer, type, expression)
					buffer.extend(';')
				end
				i := i + 1
			end
			t := agent_type.agent_result
			if t /= Void then
				t.canonical_type_mark.c_type_for_argument_in(buffer)
				buffer.append(once " R;")
			end
			buffer.append(once "};%N")
			cpp.put_string(buffer)
		ensure
			cpp.on_h
		end

feature {}
	update_argument_list (type: TYPE; target: EXPRESSION; arguments: EFFECTIVE_ARG_LIST) is
		local
			i: INTEGER
		do
			argument_list.clear_count
			agent_creation_helper.argument_list_in(argument_list, type, target, arguments)
			if ace.boost and then
				smart_eiffel.status.collecting_done and then
				not use_current(type)
			 then
				-- No need to capture `Current':
				from
					i := argument_list.upper
				until
					i < argument_list.lower
				loop
					if argument_list.item(i).is_current then
						argument_list.remove(i)
						i := argument_list.lower - 1
					else
						i := i - 1
					end
				end
			end
		end
	
	argument_list: FAST_ARRAY[EXPRESSION] is
		once
			create Result.with_capacity(8)
		end

	agent_creation_helper: AGENT_CREATION_HELPER is
		once
			create Result.make
		end

invariant
	not start_position.is_unknown

	delayed_call /= Void

end -- class AGENT_CREATION
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- SmartEiffel is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License,
-- as published by the Free Software Foundation; either version 2, or (at your option) any later version.
-- SmartEiffel is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have
-- received a copy of the GNU General Public License along with SmartEiffel; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
