-- This file is part of SmartEiffel The GNU Eiffel Compiler Tools and Libraries.
-- See the Copyright notice at the end of this file.
--
deferred class CREATE_TOOLS
	--
	-- Common root for OLD_CREATION, CREATE_INSTRUCTION and CREATE_EXPRESSION
	--

inherit
	VISITABLE
insert
	GLOBALS

feature {ANY}
	start_position: POSITION

	explicit_type: TYPE_MARK
			-- Explicit created one (optional for the CREATE_INSTRUCTION and mandatory for the CREATE_EXPRESSION).

	writable: EXPRESSION
			-- The target of the creation call (mandatory for the CREATE_INSTRUCTION and forbidden for the
			-- CREATE_EXPRESSION ).

	call: PROC_CALL
			-- Optional initialisation call if any. (The target of `call' is the `writable' entity.)

	default_create_call: BOOLEAN

	frozen side_effect_free (type: TYPE): BOOLEAN is
		do
			-- Memory allocation.
		end

	frozen safety_check (type: TYPE) is
		local
			args: like arguments
		do
			args := arguments
			if args /= Void then
				args.safety_check(type)
			end
		end

feature {ONCE_ROUTINE_POOL, PRECOMPUTABLE_ROUTINE_DETECTOR, HIDDEN_EXPRESSION_DETECTOR}
	creation_procedure (type_to_create: TYPE): E_PROCEDURE is
		require
			type_to_create /= Void
			default_create_is_inherited: call /= Void
		local
			feature_stamp: FEATURE_STAMP
		do
			-- (At this step, we are sur to have a non Void call, because `default_create' is made
			-- as the default.)
			feature_stamp := call.feature_stamp
			Result ?= feature_stamp.anonymous_feature(type_to_create)
		ensure
			Result /= Void
		end

	created_type (type: TYPE): TYPE is
		require
			smart_eiffel.status.collecting_done
		do
			if explicit_type /= Void then
				Result := explicit_type.resolve_in(type)
			else
				Result := writable.resolve_in(type)
			end
		ensure
			Result.live_type /= Void
		end

feature {CREATE_TOOLS}
	specialize_2_checks (t: TYPE) is
			-- Check the validity of the CREATION_INSTRUCTION / CREATION_EXPRESSION where `t' is the `sepcialize_2' argument.
			-- (Note because of possible redefinition checks are not wrapped inside a `is_the_validity_check_site_of' call.)
		require
			t /= Void
			call /= Void
		local
			type_to_create, writable_type: TYPE; fn: FEATURE_NAME; class_to_create: CLASS_TEXT
			feature_stamp: FEATURE_STAMP; creation_clause_list: CREATION_CLAUSE_LIST
			anonymous_feature: ANONYMOUS_FEATURE
		do
			if explicit_type /= Void then
				if explicit_type.is_formal_generic then
					-- Because `explicit_type' may be inherited:
					if start_position.class_text = explicit_type.start_position.class_text then
						error_handler.add_position(start_position)
						error_handler.append("Creation call on formal generic type (")
						error_handler.append(explicit_type.written_mark)
						error_handler.append(").")
						error_handler.print_as_fatal_error
					end
				end
				type_to_create := explicit_type.resolve_in(t)
				if writable /= Void then
					writable_type := writable.resolve_in(t)
					if not type_to_create.can_be_assigned_to(writable_type) then
						error_handler.add_position(explicit_type.start_position)
						error_handler.add_position(writable.start_position)
						error_handler.append("Cannot assign newly created object of type ")
						error_handler.append(type_to_create.name.to_string)
						error_handler.append(" into ")
						error_handler.add_expression(writable)
						error_handler.append(" which is of type ")
						error_handler.append(writable_type.name.to_string)
						error_handler.append(".")
						error_handler.print_as_fatal_error
					end
				end
			else
				type_to_create := writable.resolve_in(t)
			end
			class_to_create := type_to_create.class_text
			if class_to_create.is_deferred then
				error_handler.append("Type ")
				error_handler.append(type_to_create.name.to_string)
				error_handler.append(" is deferred. (Cannot create object.)")
				error_handler.add_position(start_position)
				error_handler.print_as_warning
			end
			creation_clause_list := class_to_create.creation_clause_list
			feature_stamp := call.feature_stamp
			anonymous_feature := feature_stamp.anonymous_feature(type_to_create)
			if anonymous_feature.result_type /= Void then
				error_handler.add_position(anonymous_feature.start_position)
				error_handler.add_position(call.start_position)
				error_handler.append(fz_ffinap)
				error_handler.print_as_fatal_error
			end
			if creation_clause_list = Void then
				if not default_create_call then
					error_handler.append("Type ")
					error_handler.append(type_to_create.name.to_string)
					error_handler.append(" has no creation list. You must use the default creation method (named %
				    %`default_create' in class GENERAL, or just omit the method name).")
					error_handler.add_position(call.start_position)
					error_handler.add_position(start_position)
					error_handler.print_as_fatal_error
				end
			else
				fn := type_to_create.get_feature_name(feature_stamp)
				if not class_to_create.has_creation_check(fn, start_position) then
					if default_create_call then
						error_handler.append("Creation clause exists for type ")
						error_handler.append(type_to_create.name.to_string)
						error_handler.append(" and the default creation procedure (see `default_create' in class GENERAL) is not allowed. %
				       %You must use one of the available creation procedure here.")
						error_handler.add_position(start_position)
						creation_clause_list.all_positions_in_error_handler
						error_handler.print_as_fatal_error
					else
						error_handler.add_position(call.feature_name.start_position)
						error_handler.add_position(fn.start_position)
						error_handler.append(fn.to_string)
						error_handler.append(" is not in the creation list of type ")
						error_handler.append(type_to_create.name.to_string)
						error_handler.append(".")
						error_handler.add_position(creation_clause_list.start_position)
						error_handler.print_as_fatal_error
					end
				end
				if fn.to_string = as_manifest_creation then
					error_handler.add_position(call.feature_name.start_position)
					error_handler.add_position(fn.start_position)
					error_handler.append("`manifest_create' is not an ordinary creation procedure. Please use the {")
					error_handler.append(type_to_create.name.to_string)
					error_handler.append(" ... << ... >> } manifest creation notation.")
					error_handler.print_as_error
				end
			end
		end

feature {MANIFEST_TUPLE}
	arguments: EFFECTIVE_ARG_LIST is
		do
			if call /= Void then
				--***PH: remove the test?
				Result := call.arguments
			end
		end

	arguments_count: INTEGER is
		do
			if call /= Void then
				--***PH: remove the test?
				Result := call.arg_count
			end
		ensure
			Result >= 0
		end

feature {}
	compile_to_jvm0 (t: TYPE_MARK) is
			-- Push the new object with default initialization.
		require
			t /= Void
		local
			dummy: INTEGER
		do
			if t.is_reference then
				t.live_type.jvm_basic_new
			else
				dummy := t.jvm_push_default
			end
		end

invariant
	not start_position.is_unknown

end -- class CREATE_TOOLS
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- SmartEiffel is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License,
-- as published by the Free Software Foundation; either version 2, or (at your option) any later version.
-- SmartEiffel is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have
-- received a copy of the GNU General Public License along with SmartEiffel; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
