-- This file is part of SmartEiffel The GNU Eiffel Compiler Tools and Libraries.
-- See the Copyright notice at the end of this file.
--
class SMART_EIFFEL
	--
	-- Singleton object to handle general purpose information about the
	-- SmartEiffel global compilation process. (This singleton is shared via
	-- the GLOBALS.`smart_eiffel' once function.)
	--

inherit
	ANY
insert
	GLOBALS

feature {ANY}
	copyright: STRING is "[
      SmartEiffel The GNU Eiffel Compiler, Eiffel tools and libraries
      Release 2.1 (Tuesday February 8th 2005) [Poirot]
      Copyright (C), 1994-2002 - INRIA - LORIA - ESIAL UHP Nancy 1 - FRANCE
      Copyright (C), 2003-2005 - INRIA - LORIA - IUT Charlemagne Nancy 2 - FRANCE
      D.COLNET, P.RIBET, C.ADRIAN, V.CROIZIER F.MERIZEN - SmartEiffel@loria.fr
      http://SmartEiffel.loria.fr

      ]"

	status: STATUS is
		once
			create Result.make
		end

	is_ready: BOOLEAN is
			-- Is all the live code already gathered
		do
			Result := status.is_safety_checking or else status.is_generating
		end

	short_or_class_check_flag: BOOLEAN
			-- True when command `short' or command `class_check' is running.

	pretty_flag: BOOLEAN
			-- True when command `pretty' is running.

	no_id: BOOLEAN
			-- True when the ids file has not to be read.

	class_text (class_name: CLASS_NAME; report_error: BOOLEAN): CLASS_TEXT is
			-- Retrieve the good one into the `class_text_dictionary' or launch the `eiffel_parser' if this class is not yet loaded.
			-- (See also `loaded_class_text'.)
		require
			class_name /= Void
		local
			cn: STRING
		do
			Result := class_text_dictionary.reference_at(class_name)
			if Result = Void then
				check
					status.is_collecting
				end
				cn := class_name.to_string
				if cn = as_none then
					create Result.none_class(class_name.hashed_name)
					echo.put_integer(class_text_count)
					echo.put_character('%T')
					echo.put_character('(')
					echo.put_string(class_name.to_string)
					echo.put_character(')')
					echo.put_character('%N')
				elseif eiffel_parser.is_running then
					error_handler.append("#1 in SMART_EIFFEL.")
					error_handler.print_as_internal_error
				else
					if ace.parser_buffer_for(class_name.hashed_name, report_error) then
						Result := eiffel_parser.analyse_class(class_name)
						check
							Result /= Void implies class_text_dictionary.has(class_name)
						end
					end
					if Result = Void and then report_error then
						error_handler.add_position(class_name.start_position)
						error_handler.append("Unable to load class %"")
						error_handler.append(class_name.to_string)
						error_handler.append("%".")
						error_handler.print_as_fatal_error
					end
				end
			end
			check
				report_error implies Result /= Void
			end
		end

	loaded_class_text (cn: CLASS_NAME): CLASS_TEXT is
			-- Retrieve the good one into the `class_text_dictionary'. When the
			-- corresponding class `cn' is not yet loaded, the `eiffel_parser'
			-- is _not_ launched and the Result is Void.
			-- (See also `class_text'.)
		do
			Result := class_text_dictionary.reference_at(cn)
		end

	same_base_feature (up_rf: RUN_FEATURE; run_time_set: RUN_TIME_SET): BOOLEAN is
			-- True when all `dynamic' features of the `run_time_set' have
			-- excately the same final name and refer exactely to the same
			-- `base_feature'.
			-- *** still valid ???
		require
			up_rf /= Void
			is_ready and run_time_set.count > 1
		local
			i: INTEGER; f: ANONYMOUS_FEATURE; dyn_rf: RUN_FEATURE; lt: LIVE_TYPE
		do
			from
				Result := True
				i := run_time_set.count
				f := up_rf.base_feature
			until
				not Result or else i = 0
			loop
				lt := run_time_set.item(i)
				dyn_rf := lt.dynamic_feature(up_rf)
				if f = dyn_rf.base_feature then
					if dyn_rf.name.to_string /= up_rf.name.to_string then
						Result := False
					end
				else
					Result := False
				end
				i := i - 1
			end
		end

	no_file_for (class_name: STRING): BOOLEAN is
		require
			string_aliaser.registered_one(class_name)
		do
			if as_none = class_name then
				Result := True
			end
		ensure
			Result = (as_none = class_name)
		end

	stupid_switch (t: TYPE_MARK; run_time_set: RUN_TIME_SET): BOOLEAN is
			-- True when `t' drives exactely to the same `t.run_type' for all
			-- element of the `run_time_set'.
			-- *** NOT CLEAR AT ALL ***
			-- Do not remember the goal of this method
			-- *** TO BE REVISITED ***
		require
			is_ready
		do
			-- ***
			Result := False -- *** SHOULD NOT BREAK ... JUST SLOW DOWN.
		end

	generating_type_used: BOOLEAN
			-- When GENERAL `generating_type' is used.

	generator_used: BOOLEAN
			-- When GENERAL `generator' is used.

	deep_twin_used: BOOLEAN
			-- When `deep_twin' support is necessary.

	scoop: BOOLEAN
			-- When some SCOOP functions are used. See scoop.html for details.

	accumulating_type: TYPE is
			-- Service provided for debugging purpose only
		do
			debug
				Result := type_any.feature_accumulator.context_type
			end
		ensure
			debugging_only: Result /= Void
		end

	unknown_feature_fatal_error (target_expression: EXPRESSION; target_type: TYPE; fn: FEATURE_NAME) is
			-- (This feature is placed here to standardize all error messages of this kind.)
		require
			target_expression /= Void
			target_type /= Void
			fn /= Void
		do
			error_handler.add_position(fn.start_position)
			if fn.is_infix_name then
				error_handler.append("Unknown infix operator %"")
				error_handler.append(fn.to_string)
				error_handler.append("%" in type ")
			elseif fn.is_prefix_name then
				error_handler.append("Unknown prefix operator %"")
				error_handler.append(fn.to_string)
				error_handler.append("%" in type ")
			else
				error_handler.append("Unknown feature `")
				error_handler.append(fn.to_string)
				error_handler.append("' in type ")
			end
			error_handler.add_type(target_type)
			error_handler.append(".")
			if target_expression.is_implicit_current then
				error_handler.append(" (For this call, the target is the implicit non written `Current' %
                                 %which is of type ")
				error_handler.add_type(target_type)
				error_handler.append(".)")
			else
				error_handler.add_position(target_expression.start_position)
				error_handler.append(" (For this call, the target ")
				error_handler.add_expression(target_expression)
				error_handler.append(" is of type ")
				error_handler.add_type(target_type)
				error_handler.append(".)")
			end
			error_handler.print_as_fatal_error
		end

feature {}
	free_inline_memo: FAST_ARRAY[INLINE_MEMO] is
		once
			create Result.with_capacity(128)
		end

feature {ANY}
	get_inline_memo: INLINE_MEMO is
		do
			if free_inline_memo.is_empty then
				create Result.make
			else
				Result := free_inline_memo.last
				free_inline_memo.remove_last
				Result.clear
			end
		ensure
			Result.cleared
		end

	dispose_inline_memo (im: INLINE_MEMO) is
		require
			im /= Void
		do
			free_inline_memo.add_last(im)
		end

feature {CREATE_INSTRUCTION, CREATE_EXPRESSION}
	deferred_dictionary_or_set_creation_check (p: POSITION; tm: TYPE_MARK): TYPE_MARK is
			-- Temporary check to help user to switch from DICTIONARY to HASHED_DICTIONARY / AVL_DICTIONARY. Here `p' is supposed to
			-- be the creation statement's position for the `tm' to check/replace. The non Void `Result' indicates a warning and is a
			-- replacement of `tm' from DICTIONARY to HASHED_DICTIONARY. (This code has been added in August 2003.) Ditto for SET.
		require
			not p.is_unknown
			tm.is_static
		local
			cn: CLASS_NAME; s: STRING
		do
			s := tm.class_text_name.to_string
			if s = as_dictionary then
				error_handler.add_position(p)
				error_handler.append("Class DICTIONARY is now a deferred class with two different implementations: %
				 %HASHED_DICTIONARY and AVL_DICTIONARY. The HASHED_DICTIONARY implementation corresponds to %
				 %the good old DICTIONARY class (i.e. hash_code based implementation). The new AVL_DICTIONARY %
				 %implementation is a tree based implementation with a different generic constraint. Please, %
				 %also note that it is not necessary to change the corresponding declaration of the target %
				 %entity and that you can select the best implementation by changing only this create %
				 %statement.")
				error_handler.print_as_warning
				create cn.make(string_aliaser.hashed_string(as_hashed_dictionary), tm.start_position)
				create {USER_GENERIC_TYPE_MARK} Result.make(cn, tm.generic_list)
				error_handler.add_position(p)
				error_handler.append("Automatic replacement of DICTIONARY with HASHED_DICTIONARY here (please update your code).")
				error_handler.print_as_warning
			elseif s = as_set then
				error_handler.add_position(p)
				error_handler.append("Class SET is now a deferred class with two different implementations: %
				 %HASHED_SET and AVL_SET. The HASHED_SET implementation corresponds to %
				 %the good old SET class (i.e. hash_code based implementation). The new AVL_SET %
				 %implementation is a tree based implementation with a different generic constraint. Please, %
				 %also note that it is not necessary to change the corresponding declaration of the target %
				 %entity and that you can select the best implementation by changing only this create %
				 %statement.")
				error_handler.print_as_warning
				create cn.make(string_aliaser.hashed_string(as_hashed_set), tm.start_position)
				create {USER_GENERIC_TYPE_MARK} Result.make(cn, tm.generic_list)
				error_handler.add_position(p)
				error_handler.append("Automatic replacement of SET with HASHED_SET here (please update your code).")
				error_handler.print_as_warning
			end
		end

feature {TYPE}
	is_checking_signature: BOOLEAN
			-- When `True', we are checking a signature conformance. In that case, we must not be hindered by "conforming" or
			-- "non-conforming" considerations.
			-- See `start_checking_signature' and `stop_checking_signature'.
			-- See also `TYPE.is_conform_to'.

feature {ANONYMOUS_FEATURE_MIXER, ANONYMOUS_FEATURE}
	start_checking_signature is
		do
			check
				not is_checking_signature
			end
			is_checking_signature := True
		ensure
			is_checking_signature
		end

	stop_checking_signature is
		do
			check
				is_checking_signature
			end
			is_checking_signature := False
		ensure
			not is_checking_signature
		end

feature {E_RESULT, ONCE_FUNCTION}
	vffd7_fatal_error (sp: POSITION) is
		do
			error_handler.add_position(sp)
			error_handler.append("Result type of a once function must not involve formal generic names nor anchored types (VFFD.8).")
			error_handler.print_as_fatal_error
		end

feature {LIVE_TYPE, CREATE_TOOLS}
	set_scoop is
		do
			scoop := True
			if ace.profile then
				error_handler.append(once "-profile and SCOOP don't work well together. I suggest you not using -profile.%N")
				error_handler.print_as_warning
			end
		end

feature {GC_HANDLER, C_PRETTY_PRINTER}
	root_procedure: RUN_FEATURE_3
			-- The root procedure of the system to compile.

feature {PRETTY}
	class_text_for_pretty (file_path: STRING; class_name: CLASS_NAME): CLASS_TEXT is
		require
			file_exists(file_path)
			class_name /= Void
		do
			check
				not class_text_dictionary.has(class_name)
			end
			parser_buffer.load_file(file_path)
			parser_buffer.set_cluster(create {CLUSTER}.for_pretty)
			Result := eiffel_parser.analyse_class(class_name)
		end

	remove_loaded_class (ct: CLASS_TEXT) is
			-- Removed the already loaded `ct' in order to allow `pretty' to parse again the generated file to check that the newly
			-- created file is syntactically correct.
		require
			ct /= Void
		local
			name: CLASS_NAME
		do
			name := ct.name
			check
				class_text_dictionary.has(name)
			end
			class_text_dictionary.remove(name)
		end

feature {CLASS_TEXT}
	add_class_text (ct: CLASS_TEXT) is
		require
			ct /= Void
		local
			name: CLASS_NAME
		do
			name := ct.name
			check
				not class_text_dictionary.has(name)
			end
			class_text_dictionary.add(ct, name)
		ensure
			class_text_dictionary.has(ct.name)
		end

feature {FINDER}
	find_path_for (arg: STRING): STRING is
		local
			hs: HASHED_STRING
		do
			hs := string_aliaser.hashed_string(arg)
			if ace.parser_buffer_for(hs, True) then
				Result := parser_buffer.path
				parser_buffer.release
			end
		end

feature {CLASS_CHECKER}
	tuple_related_classes_in (bc: CLASS_TEXT; parent_list: FAST_ARRAY[TYPE]) is
		require
			parent_list.is_empty
		local
			ct, tuple_ct: CLASS_TEXT; tuple_type: TYPE; cn: CLASS_NAME; stop: BOOLEAN; i, tuple_index: INTEGER
		do
			from
				i := 1
			until
				stop
			loop
				if i > class_text_dictionary.count then
					if tuple_type = Void then
						-- No more TUPLE class text.
						stop := True
					else
						tuple_type := Void
					end
					i := 1
				else
					ct := class_text_dictionary.item(i)
					cn := ct.name
					if cn.is_tuple_related then
						if ct.formal_generic_list = Void then
							if tuple_index = 0 then
								tuple_ct := ct
							end
						elseif ct.formal_generic_list.count = tuple_index then
							tuple_ct := ct
						end
						if tuple_ct /= Void then
							tuple_type := tuple_ct.declaration_type_of_like_current
							parent_list.add_last(tuple_type)
							tuple_index := tuple_index + 1
							i := 1
							tuple_ct := Void
							tuple_type := Void
						end
					end
					i := i + 1
				end
			end
		end

feature {CLASS_CHECKER, EXTERNAL_TOOL}
	set_short_or_class_check_flag is
		do
			short_or_class_check_flag := True
			set_no_id
		end

feature {PRETTY}
	set_pretty_flag is
		do
			pretty_flag := True
			set_no_id
		end

feature {}
	set_no_id is
		do
			no_id := True
		end

feature {ANY}
	class_text_count: INTEGER is
			-- Total number of class text actually loaded.
		do
			Result := class_text_dictionary.count
		end

	compile_to_c is
			-- Produce C code for `root_class'/`procedure'.
		local
			root_class, procedure: STRING; lt: LIVE_TYPE; run_count, i: INTEGER; gc_flag: BOOLEAN
			hashed_string: HASHED_STRING
		do
			initialize_general_any_tuple
			root_class := ace.root_class_name.to_string
			procedure := ace.root_procedure_name
			get_started(root_class, procedure)
			if nb_errors = 0 then
				check
					root_procedure /= Void
				end
				cpp.get_started
				check
					is_ready
				end
				cpp.swap_on_h
				gc_flag := not gc_handler.is_off -- ---------------------------------------------------------
				if ace.boost then
					cpp.put_string_on_h(once "#define SE_BOOST 1%N")
				end
				if ace.sedb then
					cpp.put_string_on_h(once "#define SE_SEDB 1%N")
				end
				if gc_flag then
					cpp.put_string_on_h(once "#define SE_GC_LIB 1%N")
				end
				if exceptions_handler.used then
					cpp.put_string_on_h(once "#define SE_EXCEPTIONS 1%N")
				end
				if scoop then
					cpp.put_string_on_h(once "#define SE_SCOOP 1%N")
				end
				-- ---------------------------------------------------------
				if scoop then
					cpp.sys_runtime_h_and_c(once "scoop")
					cpp.sys_runtime_h_and_c(once "scoop_thread")
					system_tools.add_lib_scoop
				end
				-- ---------------------------------------------------------
				if ace.profile then
					cpp.sys_runtime_h_and_c(once "profile")
				end
				-- ---------------------------------------------------------
				from
					cpp.put_comment_line(once "C Header Pass 1 :")
					agent_pool.c_header_pass1
					i := live_type_map.upper
				until
					i < 0
				loop
					lt := live_type_map.item(i)
					if lt.at_run_time then
						run_count := run_count + 1
					end
					lt.c_header_pass1
					i := i - 1
				end
				-- ---------------------------------------------------------
				from
					cpp.put_comment_line(once "C Header Pass 2 :")
					i := live_type_map.upper
				until
					i < 0
				loop
					lt := live_type_map.item(i)
					lt.c_header_pass2
					i := i - 1
				end
				-- ---------------------------------------------------------
				from
					cpp.put_comment_line(once "C Header Pass 3 :")
					i := live_type_map.upper
				until
					i < 0
				loop
					lt := live_type_map.item(i)
					lt.c_header_pass3
					i := i - 1
				end
				-- ---------------------------------------------------------
				from
					cpp.put_comment_line(once "C Header Pass 4 :")
					i := live_type_map.upper
				until
					i < 0
				loop
					lt := live_type_map.item(i)
					lt.c_header_pass4
					i := i - 1
				end
				hashed_string := string_aliaser.hashed_string(as_native_array_character)
				if not type_dictionary.has(hashed_string) then
					-- Force definition of T9 and T7:
					cpp.put_string(once "typedef T3* T9;%N")
					if ace.no_check then
						cpp.put_c_function(once "void se_prinT9(FILE* file, T9*o)", once "fprintf(file, %"NATIVE_ARRAY[STRING]#%%p\n%",(void*)*o);")
					end
				end
				manifest_string_pool.c_define1(string_at_run_time)
				cpp.customize_runtime(collected_plug_in)
				-- ---------------------------------------------------------
				if gc_flag then
					gc_handler.define1
				end
				-- ---------------------------------------------------------
				compile_routines
			   cpp.cecil_define
				agent_pool.customize_c_runtime_2
				-- ---------------------------------------------------------
				if ace.sedb then
					prepare_introspection
				end
				-- ---------------------------------------------------------
				cpp.split_c_file_padding_here
				manifest_string_pool.c_define2(string_at_run_time)
				compile_registered_for_c_define
				address_of_pool.c_define
				if gc_flag then
					gc_handler.define2
				end
				cpp.define_main(root_procedure)
				cpp.split_c_file_padding_here
				manifest_array_pool.c_define -- *** To be removed together with old << >> *** (Dom Oct 28th 2004) ***
				manifest_generic_pool.c_define
				cpp.split_c_file_padding_here
				switch_collection.c_define
				cpp.split_c_file_padding_here
				if scoop then
					separate_tools.generate_scoop_functions(live_type_map)
				end
            assignment_test_pool.c_define_assignment_test_functions(live_type_map)
				debug
					echo.put_string(once "Very Final magic_count : ")
					echo.put_integer(magic_count)
					echo.put_character('%N')
				end
				cpp.write_make_file
				cpp.echo_information
				very_last_information
			else
				error_handler.append("Cannot produce C code.")
				error_handler.print_as_error
			end
		end

	compile_to_jvm is
			-- Produce Java Byte Code for `root_class'/`procedure'.
		local
			root_class, procedure: STRING; lt: LIVE_TYPE; run_count, i: INTEGER
		do
			initialize_general_any_tuple
			root_class := ace.root_class_name.to_string
			procedure := ace.root_procedure_name
			get_started(root_class, procedure)
			if scoop then
				not_yet_implemented
			end
			if nb_errors = 0 then
				jvm.prepare_output_directory
				from
					i := live_type_map.upper
				until
					i < 0
				loop
					lt := live_type_map.item(i)
					if lt.at_run_time then
						run_count := run_count + 1
						lt.compile_to_jvm
					end
					i := i - 1
				end
				echo.print_count(once "Used Type", run_count)
				agent_pool.customize_jvm_runtime
				jvm.write_jvm_root_class
				jvm.write_main_class(root_procedure)
				very_last_information
			else
				error_handler.append(once "Cannot produce Java Byte Code.")
				error_handler.print_as_error
			end
		end

feature {}
	class_text_dictionary: DICTIONARY[CLASS_TEXT, CLASS_NAME] is
			-- When looking for a CLASS_TEXT using the name of the class text
			-- (ie. FOO[BAR] is stored at key "FOO").
		once
			create {HASHED_DICTIONARY[CLASS_TEXT, CLASS_NAME]} Result.with_capacity(1024)
		end

	type_dictionary: DICTIONARY[TYPE, HASHED_STRING] is
			-- When looking for a TYPE using it's name (ie. FOO[BAR] is stored at key whith name "FOO[BAR]").
		once
			create {HASHED_DICTIONARY[TYPE, HASHED_STRING]} Result.with_capacity(1024)
		end

feature {ANY}
	magic_count: INTEGER
			-- Grow each time a new run class is added, each time a new class is
			-- loaded, each time a new feature is checked, each time another
			-- expression is optimized, etc. ...
			-- Thus when `magic_count' stops growing, we are really ready :-).

	magic_count_increment is
		do
			magic_count := magic_count + 1
		end

feature {GC_HANDLER}
	live_type_map: FAST_ARRAY[LIVE_TYPE] is
			-- *** an attribute created in init_phase would be more
			-- *** efficient than this once function. (in default_create)
			--***(PR 29/05/04) this once should be precomputed
		once
			create Result.with_capacity(2048)
		end

feature {CALL_PROC_CALL, WRITABLE_ATTRIBUTE_NAME, MANIFEST_STRING_POOL, MANIFEST_ARRAY, CREATION_CLAUSE, ADDRESS_OF, ONCE_FUNCTION, MANIFEST_GENERIC}
	collect (type: TYPE; fs: FEATURE_STAMP; at_run_time: BOOLEAN): TYPE is
			-- The `Result' is not Void when `fs' is actually a function (see ensure).
		require
			fs.has_anonymous_feature_for(type)
		local
			lt: LIVE_TYPE; af: ANONYMOUS_FEATURE; tm: TYPE_MARK; tmp_fs: FEATURE_STAMP
		do
			--***PH(03/03/04) To be removed
			lt := collect_one_type(type, at_run_time)
			--***PH(03/03/04) should be "af := fs.anonymous_feature(type)",
			--but in order to quickly solve a hard problem I added one
			--extra resolution.
			tmp_fs := fs.resolve_dynamic_binding_for(type)
			af := tmp_fs.anonymous_feature(type)
			tm := af.result_type
			if tm /= Void then
				Result := tm.resolve_in(type)
			end
			lt.collect(tmp_fs)
		ensure
			fs.anonymous_feature(type).result_type /= Void implies Result = fs.anonymous_feature(type).result_type.resolve_in(type)
		end

feature {LOCAL_VAR_LIST, ANONYMOUS_FEATURE}
	collect_local_expanded (type: TYPE) is
			-- Make live the given `type'.
		require
			type.is_expanded
		local
			lt: LIVE_TYPE
		do
			lt := collect_one_type(type, True)
			-- *** May be we have to launch here the collect on the
			-- corresponding creation procedure without argument ?
			-- (Dom. 8/9/2003)
		ensure
			type.live_type /= Void
		end

feature {BASE_TYPE_CONSTANT}
	collect_constant (type: TYPE) is
			-- Make live the given `type' (of a constant)
		require
			type /= Void
		local
			lt: LIVE_TYPE
		do
			lt := collect_one_type(type, True)
		ensure
			type.live_type /= Void
		end

feature {EXTERNAL_FUNCTION}
	collect_external (type: TYPE) is
			-- Make live the given type (of an external function)
		require
			type /= Void
		local
			lt: LIVE_TYPE
		do
			lt := collect_one_type(type, True)
		ensure
			type.live_type /= Void
		end

feature {ASSERTION}
	collect_assertion (type: TYPE) is
			-- Make live the given type (of an assertion)
		require
			type.is_boolean
		local
			lt: LIVE_TYPE
		do
			lt := collect_one_type(type, True)
		ensure
			type.live_type /= Void
		end

feature {AGENT_CREATION, ADDRESS_OF}
	collect_create (type: TYPE) is
		require
			type /= Void
		local
			lt: LIVE_TYPE
		do
			lt := collect_one_type(type, True)
		ensure
			type.live_type.at_run_time
		end

feature {LIVE_TYPE}
	collect_generic (type: TYPE) is
		require
			type /= Void
		local
			lt: LIVE_TYPE
		do
			lt := collect_one_type(type, False)
		ensure
			type.live_type /= Void
		end

	collect_se_atexit (type: TYPE) is
		require
			type /= Void
			no_extra_computation: se_atexit_stamp = Void
		local
			dummy: TYPE
		do
			se_atexit_stamp := type.feature_stamp_of(se_atexit_name)
			dummy := collect(type, se_atexit_stamp, False)
			se_atexit_id := type.live_type.id
		ensure
			se_atexit_stamp /= Void
		end

feature {CODE_PRINTER}
	se_atexit_stamp: FEATURE_STAMP

	se_atexit_id: INTEGER

feature {}
	collected_external_functions: FAST_ARRAY[NON_VOID_NO_DISPATCH] is
		once
			create Result.with_capacity(3)
		end

feature {NON_VOID_NO_DISPATCH}
	collect_external_function (non_void_no_dispatch: NON_VOID_NO_DISPATCH; fs: FEATURE_STAMP; type: TYPE) is
		require
			non_void_no_dispatch /= Void
			type /= Void
			fs /= Void
		local
			dummy: TYPE
		do
			if not collected_external_functions.fast_has(non_void_no_dispatch) then
				dummy := collect(type, fs, False)
				collected_external_functions.add_last(non_void_no_dispatch)
			end
		end

feature {ONCE_ROUTINE_POOL}
	collect_precomputable (type: TYPE; fs: FEATURE_STAMP) is
		require
			type /= Void
			fs /= Void
		local
			dummy: TYPE
		do
			dummy := collect(type, fs, False)
		end

feature {CECIL_ENTRY}
	collect_cecil_entry (type: TYPE; fs: FEATURE_STAMP) is
		require
			type /= Void
			fs /= Void
		local
			dummy: TYPE
		do
			dummy := collect(type, fs, False)
		end

feature {CALL, PROC_CALL, PRECURSOR_CALL, AGENT_CREATION}
	argument_count_check (call_site: POSITION; af: ANONYMOUS_FEATURE; actual_args: EFFECTIVE_ARG_LIST) is
			-- Check that the number of arguments of `af' is compatible with `actual_args'. Only the number of
			-- arguments is checked here. (The complete conformance test is performed in EFFECTIVE_ARG_LIST.)
		require
			not call_site.is_unknown
			af /= Void
		local
			formal_args: FORMAL_ARG_LIST; actual, formal: INTEGER
		do
			formal_args := af.arguments
			if formal_args /= Void then
				if actual_args /= Void then
					formal := formal_args.count
					actual := actual_args.count
					if actual /= formal then
						error_handler.append("The feature called has ")
						error_handler.append_integer(formal)
						error_handler.append(" formal argument")
						if formal > 1 then
							error_handler.extend('s')
						end
						error_handler.append(" while the actual argument list has ")
						error_handler.append_integer(actual)
						error_handler.append(" argument")
						if actual > 1 then
							error_handler.extend('s')
						end
						error_handler.append(".")
						error_handler.add_position(formal_args.start_position)
						error_handler.add_position(actual_args.start_position)
						error_handler.print_as_fatal_error
					end
				else
					formal := formal_args.count
					error_handler.append("The feature called has ")
					error_handler.append_integer(formal)
					error_handler.append(" formal argument")
					if formal > 1 then
						error_handler.extend('s')
					end
					error_handler.append(" while there is no actual argument list in the call.")
					error_handler.add_position(formal_args.start_position)
					error_handler.add_position(call_site)
					error_handler.print_as_fatal_error
				end
			elseif actual_args /= Void then
				actual := actual_args.count
				error_handler.append("The feature called has no formal argument while the actual argument list has ")
				error_handler.append_integer(actual)
				error_handler.append(" argument")
				if actual > 1 then
					error_handler.extend('s')
				end
				error_handler.append(".")
				error_handler.add_position(af.start_position)
				error_handler.add_position(actual_args.start_position)
				error_handler.print_as_fatal_error
			end
		end

feature {AGENT_POOL, CREATE_TOOLS, ASSIGNMENT, ASSIGNMENT_ATTEMPT, CREATE_WRITABLE, EFFECTIVE_ARG_LIST}
	collect_one_type (type: TYPE; at_run_time: BOOLEAN): LIVE_TYPE is
			-- Make live the given `type'.
		require
			type /= Void
		local
			rt: BOOLEAN
		do
			Result := type.live_type
			if Result = Void then
				create Result.make(type)
				check
					not type.is_expanded implies not Result.at_run_time
					type_dictionary.reference_at(type.name).live_type = Result
				end
			end
			rt := at_run_time or else Result.is_expanded and then Result.name.to_string /= as_integer_general
			if rt and then not Result.at_run_time then
				debug
					if not at_run_time then
						-- sedb_breakpoint -- type will become alive, is it really necessary?
						-- *** I have noticed that it is often NATIVE_ARRAY here ...
						-- *** NATIVE_ARRAY must be made alive when {NATIVE_ARRAY}.malloc is collected !!
						-- *** (Dom. april 22th 2004) ***
					end
				end
				Result.set_at_run_time
			end
		ensure
			is_alive: type.live_type /= Void
		end

feature {ANONYMOUS_FEATURE, RUN_FEATURE, ARGUMENT_NAME2, LOCAL_NAME2, E_RESULT, E_OLD}
	context_feature: ANONYMOUS_FEATURE is
		do
			Result := context_feature_stack.last
		end

feature {ANONYMOUS_FEATURE, RUN_FEATURE}
	push_context (af: ANONYMOUS_FEATURE) is
		do
			context_feature_stack.add_last(af)
		end

	pop_context (af: ANONYMOUS_FEATURE) is
		require
			context_feature = af
		do
			context_feature_stack.remove_last
		end

feature {}
	context_feature_stack: FAST_ARRAY[ANONYMOUS_FEATURE] is
			-- We really need a stack. As an example, it is necessary to follow the calling graph during
			-- collect.
		once
			create Result.with_capacity(4)
		end

feature {CLASS_TEXT, KERNEL_EXPANDED_TYPE_MARK}
	live_type_for (cn: CLASS_NAME): LIVE_TYPE is
			-- Assume `cn' is a simple non generic class.
			-- obsolete "use collect"
		local
			ct: CLASS_TEXT; t: TYPE; type_class: CLASS_TYPE_MARK
		do
			echo.put_string(once "*** {SMART_EIFFEL}.live_type_for is still used!!! ***%N")
			t := type_dictionary.reference_at(cn.hashed_name)
			if t /= Void then
				Result := t.live_type
			end
			if Result = Void then
				ct := class_text_dictionary.reference_at(cn)
				if ct /= Void then
					create type_class.make(ct.name)
					Result := type_class.live_type
				else
					error_handler.append("#2 in SMART_EIFFEL for: ")
					error_handler.append(cn.to_string)
					error_handler.print_as_internal_error
				end
			end
		ensure
			Result /= Void
		end

feature {LIVE_TYPE}
	is_tagged (lt: LIVE_TYPE): BOOLEAN is
		require
			is_ready
			lt.at_run_time
			lt.is_reference
			ace.boost
		local
			i: INTEGER; run_time_set: RUN_TIME_SET
		do
			from
				i := live_type_map.upper
			until
				Result or else i < 0
			loop
				run_time_set := live_type_map.item(i).run_time_set
				if run_time_set.count = 0 then
				elseif run_time_set.has(lt) then
					Result := run_time_set.count > 1
				end
				i := i - 1
			end
		end

	class_used (name: STRING): CLASS_TEXT is
			-- The class text when used, or Void when this class is not in the live code.
		require
			aliased: string_aliaser.registered_one(name)
		local
			cn: CLASS_NAME
		do
			create cn.unknown_position(string_aliaser.hashed_string(name))
			Result := class_text_dictionary.reference_at(cn)
		end

feature {LOCAL_NAME2, EFFECTIVE_ROUTINE}
	specializing_feature_local_var_list: LOCAL_VAR_LIST

feature {EFFECTIVE_ROUTINE}
	set_specializing_feature_variables (lvl: LOCAL_VAR_LIST) is
		do
			specializing_feature_local_var_list := lvl
		end

feature {}
	type_to_be_created: HASHED_STRING

	create_type (static_type: TYPE_MARK): TYPE is
			-- Create a type. Type creation is locked until `register_type' has been called.
		require
			static_type.is_static
			lock_type_creation(static_type.written_name)
			not has_type(static_type)
		do
			create Result.make(static_type)
		ensure
			Result /= Void
			has_type(static_type)
		end

	lock_type_creation (t: like type_to_be_created): BOOLEAN is
		do
			Result := type_to_be_created = Void
			type_to_be_created := t
		ensure
			Result
		end

feature {}
	type_any_memory: TYPE

	type_none_memory: TYPE

	type_general_memory: TYPE

	type_boolean_memory: TYPE

	type_character_memory: TYPE

	type_string_memory: TYPE

	type_pointer_memory: TYPE

	type_integer_8_memory: TYPE

	type_integer_16_memory: TYPE

	type_integer_memory: TYPE

	type_integer_64_memory: TYPE

	type_real_32_memory: TYPE

	type_real_memory: TYPE

	type_real_extended_memory: TYPE
	
feature {ANY} -- To get a TYPE:
	--*** TYPE creation can be quite recursive, so these cannot be once functions <FM-14/10/2004>

	type_any: TYPE is
		local
			hs: HASHED_STRING; unknown_position: POSITION
		do
			if type_any_memory = Void then
				hs := string_aliaser.hashed_string(as_any)
				type_any_memory := type_dictionary.reference_at(hs)
				if type_any_memory = Void then
					type_any_memory := get_type(create {ANY_TYPE_MARK}.make(unknown_position))
				end
			end
			Result := type_any_memory
		ensure
			type_any_memory /= Void
			Result /= Void
		end

	type_none: TYPE is
		local
			hs: HASHED_STRING; unknown_position: POSITION
		do
			if type_none_memory = Void then
				hs := string_aliaser.hashed_string(as_none)
				type_none_memory := type_dictionary.reference_at(hs)
				if type_none_memory = Void then
					type_none_memory := get_type(create {NONE_TYPE_MARK}.make(unknown_position))
				end
			end
			Result := type_none_memory
		ensure
			type_none_memory /= Void
			Result /= Void
		end

	type_general: TYPE is
		local
			hs: HASHED_STRING; cn: CLASS_NAME; unknown_position: POSITION
		do
			if type_general_memory = Void then
				hs := string_aliaser.hashed_string(as_general)
				type_general_memory := type_dictionary.reference_at(hs)
				if type_general_memory = Void then
					create cn.make(hs, unknown_position)
					type_general_memory := get_type(create {CLASS_TYPE_MARK}.make(cn))
				end
			end
			Result := type_general_memory
		ensure
			type_general_memory /= Void
			Result /= Void
		end

	type_boolean: TYPE is
		local
			hs: HASHED_STRING; unknown_position: POSITION
		do
			if type_boolean_memory = Void then
				hs := string_aliaser.hashed_string(as_boolean)
				type_boolean_memory := type_dictionary.reference_at(hs)
				if type_boolean_memory = Void then
					type_boolean_memory := get_type(create {BOOLEAN_TYPE_MARK}.make(unknown_position))
				end
			end
			Result := type_boolean_memory
		ensure
			type_boolean_memory /= Void
			Result /= Void
		end

	type_character: TYPE is
		local
			hs: HASHED_STRING; unknown_position: POSITION
		do
			if type_character_memory = Void then
				hs := string_aliaser.hashed_string(as_character)
				type_character_memory := type_dictionary.reference_at(hs)
				if type_character_memory = Void then
					type_character_memory := get_type(create {CHARACTER_TYPE_MARK}.make(unknown_position))
				end
			end
			Result := type_character_memory
		ensure
			type_character_memory /= Void
			Result /= Void
		end

	type_string: TYPE is
		local
			hs: HASHED_STRING; unknown_position: POSITION
		do
			if type_string_memory = Void then
				hs := string_aliaser.hashed_string(as_string)
				type_string_memory := type_dictionary.reference_at(hs)
				if type_string_memory = Void then
					type_string_memory := get_type(create {STRING_TYPE_MARK}.make(unknown_position))
				end
			end
			Result := type_string_memory
		ensure
			type_string_memory /= Void
			Result /= Void
		end

	type_pointer: TYPE is
		local
			hs: HASHED_STRING; unknown_position: POSITION
		do
			if type_pointer_memory = Void then
				hs := string_aliaser.hashed_string(as_pointer)
				type_pointer_memory := type_dictionary.reference_at(hs)
				if type_pointer_memory = Void then
					type_pointer_memory := get_type(create {POINTER_TYPE_MARK}.make(unknown_position))
				end
			end
			Result := type_pointer_memory
		ensure
			type_pointer_memory /= Void
			Result /= Void
		end

	type_integer_8: TYPE is
		local
			hs: HASHED_STRING; unknown_position: POSITION
		do
			if type_integer_8_memory = Void then
				hs := string_aliaser.hashed_string(as_integer_8)
				type_integer_8_memory := type_dictionary.reference_at(hs)
				if type_integer_8_memory = Void then
					type_integer_8_memory := get_type(create {INTEGER_TYPE_MARK}.integer_8(unknown_position))
				end
			end
			Result := type_integer_8_memory
		ensure
			type_integer_8_memory /= Void
			Result /= Void
		end

	type_integer_16: TYPE is
		local
			hs: HASHED_STRING; unknown_position: POSITION
		do
			if type_integer_16_memory = Void then
				hs := string_aliaser.hashed_string(as_integer_16)
				type_integer_16_memory := type_dictionary.reference_at(hs)
				if type_integer_16_memory = Void then
					type_integer_16_memory := get_type(create {INTEGER_TYPE_MARK}.integer_16(unknown_position))
				end
			end
			Result := type_integer_16_memory
		ensure
			type_integer_16_memory /= Void
			Result /= Void
		end

	type_integer: TYPE is
			-- Note: internally, no type for INTEGER_32.
		local
			hs: HASHED_STRING; unknown_position: POSITION
		do
			if type_integer_memory = Void then
				hs := string_aliaser.hashed_string(as_integer)
				type_integer_memory := type_dictionary.reference_at(hs)
				if type_integer_memory = Void then
					type_integer_memory := get_type(create {INTEGER_TYPE_MARK}.integer(unknown_position))
				end
			end
			Result := type_integer_memory
		ensure
			type_integer_memory /= Void
			Result /= Void
		end

	type_integer_64: TYPE is
			-- Note: internally, no type for INTEGER_32.
		local
			hs: HASHED_STRING; unknown_position: POSITION
		do
			if type_integer_64_memory = Void then
				hs := string_aliaser.hashed_string(as_integer_64)
				type_integer_64_memory := type_dictionary.reference_at(hs)
				if type_integer_64_memory = Void then
					type_integer_64_memory := get_type(create {INTEGER_TYPE_MARK}.integer_64(unknown_position))
				end
			end
			Result := type_integer_64_memory
		ensure
			type_integer_64_memory /= Void
			Result /= Void
		end

	type_real_32: TYPE is
		local
			hs: HASHED_STRING; unknown_position: POSITION
		do
			if type_real_32_memory = Void then
				hs := string_aliaser.hashed_string(as_real_32)
				type_real_32_memory := type_dictionary.reference_at(hs)
				if type_real_32_memory = Void then
					type_real_32_memory := get_type(create {REAL_TYPE_MARK}.real_32(unknown_position))
				end
			end
			Result := type_real_32_memory
		ensure
			type_real_32_memory /= Void
			Result /= Void
		end

	type_real: TYPE is
			-- Note: internally, no type for REAL_64.
		local
			hs: HASHED_STRING; unknown_position: POSITION
		do
			if type_real_memory = Void then
				hs := string_aliaser.hashed_string(as_real)
				type_real_memory := type_dictionary.reference_at(hs)
				if type_real_memory = Void then
					type_real_memory := get_type(create {REAL_TYPE_MARK}.real(unknown_position))
				end
			end
			Result := type_real_memory
		ensure
			type_real_memory /= Void
			Result /= Void
		end

	type_real_extended: TYPE is
		local
			hs: HASHED_STRING; unknown_position: POSITION
		do
			if type_real_extended_memory = Void then
				hs := string_aliaser.hashed_string(as_real_extended)
				type_real_extended_memory := type_dictionary.reference_at(hs)
				if type_real_extended_memory = Void then
					type_real_extended_memory := get_type(create {REAL_TYPE_MARK}.real_extended(unknown_position))
				end
			end
			Result := type_real_extended_memory
		ensure
			type_real_extended_memory /= Void
			Result /= Void
		end

	unlock_type_creation (t: like type_to_be_created): BOOLEAN is
		do
			Result := type_to_be_created = t
			type_to_be_created := Void
		ensure
			Result
		end

	has_type (static_type: TYPE_MARK): BOOLEAN is
		require
			static_type.is_static
		do
			Result := type_dictionary.has(static_type.written_name)
		end

	get_type (static_type: TYPE_MARK): TYPE is
		require
			static_type.is_static
		local
			name: HASHED_STRING; tm: TYPE_MARK; gl: ARRAY[TYPE_MARK]; i: INTEGER
		do
			name := static_type.written_name
			Result := type_dictionary.reference_at(name)
			if Result = Void then
				-- Before creating the TYPE of `static_type', we must be sure that it's generic 
				-- argument, if any, are created first:
				if static_type.is_generic then
					from
						gl := static_type.generic_list
						i := gl.lower
					until
						i > gl.upper
					loop
						tm := gl.item(i)
						if type_dictionary.reference_at(tm.written_name) = Void then
							-- We must create the corresponding generic argument:
							Result := get_type(tm)
						end
						i := i + 1
					end
				end
				-- Because it is possible that the previous computation actually resulted in creating the 
				-- type of `static_type' itself:
				Result := type_dictionary.reference_at(name)
				if Result = Void then
					Result := create_type(static_type)
				end
			end
		ensure
			Result /= Void
			has_type(static_type)
		end

	get_array_type (gen_type: TYPE; pos: POSITION): TYPE is
			-- Returns the ARRAY[gen_type] type
		local
			name: STRING; hs: HASHED_STRING; atm: ARRAY_TYPE_MARK
		do
			name := once ""
			name.copy(once "ARRAY[")
			name.append(gen_type.name.to_string)
			name.extend(']')
			hs := string_aliaser.hashed_string(name)
			Result := type_dictionary.reference_at(hs)
			if Result = Void then
				create atm.make(pos, gen_type.canonical_type_mark)
				Result := atm.type
			end
		ensure
			Result /= Void
		end

feature {CODE_PRINTER, LIVE_TYPE}
	dispatch_special (type: TYPE; feature_stamp: FEATURE_STAMP): STRING is
			-- For some very special cases, we must not use switch. A non Void Result indicates that
			-- we must not use a switch for this feature.
		local
			fs: FEATURE_STAMP
		do
			-- I am not really happy with this function which is called
			-- very often during code generation. May be we'll have
			-- another solution with the new inlined dispatch
			-- strategy... for the next release.
			-- *** (Dom. june 18th 2004) ***
			fs := general_c_inline_c_feature_stamp
			if feature_stamp = fs or else feature_stamp = fs.resolve_dynamic_binding_for(type) then
				Result := as_c_inline_c
			end
			if Result = Void then
				fs := general_c_inline_h_feature_stamp
				if feature_stamp = fs or else feature_stamp = fs.resolve_dynamic_binding_for(type) then
					Result := as_c_inline_h
				end
			end
		end

feature {}
	general_c_inline_c_feature_stamp: FEATURE_STAMP is
		once
			Result := feature_stamp_of(as_general, as_c_inline_c)
		end

	general_c_inline_h_feature_stamp: FEATURE_STAMP is
		once
			Result := feature_stamp_of(as_general, as_c_inline_h)
		end

	feature_stamp_of (class_name, feature_name: STRING): FEATURE_STAMP is
		require
			string_aliaser.registered_one(class_name)
			string_aliaser.registered_one(feature_name)
			string_aliaser.hashed_string(feature_name).is_simple_feature_name
		local
			hs: HASHED_STRING; tm: TYPE
		do
			hs := string_aliaser.hashed_string(class_name)
			tm := type_dictionary.reference_at(hs)
			Result := tm.feature_stamp_of(string_aliaser.hashed_string(feature_name))
		ensure
			Result /= Void
		end

feature {LIVE_TYPE}
	memory_dispose_stamp: FEATURE_STAMP is
			-- Feature stamp of {DISPOSABLE}.dispose
		require
			status.is_collecting -- Because the call can make DISPOSABLE live
		local
			hs: HASHED_STRING; cn: CLASS_NAME; tm: TYPE
		once
			hs := string_aliaser.hashed_string(as_disposable)
			tm := type_dictionary.reference_at(hs)
			if tm = Void then
				create cn.unknown_position(hs)
				tm := get_type(create {CLASS_TYPE_MARK}.make(cn))
				check
					type_dictionary.has(hs)
				end
			end
			if tm.has_simple_feature_name(dispose_name) then
				Result := tm.feature_stamp_of(dispose_name)
			else
				error_handler.append("Class DISPOSABLE must have a `dispose' feature.")
				error_handler.add_position(tm.class_text.name.start_position)
				error_handler.print_as_fatal_error
			end
		ensure
			Result /= Void
		end

feature {MANIFEST_GENERIC}
	manifest_creation_name: FEATURE_NAME is
			-- feature name of {GENERAL}.manifest_creation
		require
			status.is_collecting -- Overkill ? <FM-17/12/2004>
		local
			t: TYPE; hs: HASHED_STRING
		once
			t := type_general
			hs := string_aliaser.hashed_string(as_manifest_creation)
			if t.has_simple_feature_name(hs) then
				Result := t.feature_stamp_of(hs).anonymous_feature(t).first_name
			else
				error_handler.append("Class GENERAL must have a `manifest_creation' feature.")
				error_handler.add_position(t.class_text.name.start_position)
				error_handler.print_as_fatal_error
			end
		ensure
			Result /= Void
		end

feature {TYPE}
	register_type (t: TYPE) is
		require
			t /= Void
			unlock_type_creation(t.name)
		do
			type_dictionary.add(t, t.name)
		end

feature {LIVE_TYPE}
	register_live_type (lt: LIVE_TYPE) is
		require
			lt /= Void
		local
			min, max, n: INTEGER; stop: BOOLEAN; slt, s: STRING
		do
			if ace.boost or else live_type_map.is_empty then
				live_type_map.add_last(lt)
			else
				-- Insert it sorted by class name (useful to help recompiling the least possible C chunks)
				-- (Binary algorithm taken and adapted from ABSTRACT_SORTER.index_of)
				from
					slt := lt.class_text_name.to_string
					min := live_type_map.lower
					max := live_type_map.upper
				variant
					max - min
				until
					stop
				loop
					if min = max then
						stop := True
						if live_type_map.item(min).class_text_name.to_string < slt then
							n := min + 1
						else
							n := min
						end
					else
						check
							min < max
						end
						n := (min + max) #// 2
						s := live_type_map.item(n).class_text_name.to_string
						if s < slt then
							min := n + 1
						else
							max := n
						end
					end
				end
				live_type_map.add(lt, n)
			end
		end

feature {CLASS_CHECK}
	very_last_information is
		do
			assignment_handler.echo_information
			status.info
			eiffel_parser.show_nb_warnings
			eiffel_parser.show_nb_errors
			echo.put_string(fz_02)
		end

feature {}
	generic_formal_arguments_check is
		local
			i: INTEGER
		do
			from
				i := 1
			until
				i > class_text_dictionary.count
			loop
				class_text_dictionary.item(i).generic_formal_arguments_check
				i := i + 1
			end
		end

feature {ID_PROVIDER}
	id_extra_information (tfw: TEXT_FILE_WRITE; name: HASHED_STRING) is
		local
			ct: CLASS_TEXT; lt: LIVE_TYPE; path: STRING; cn: CLASS_NAME; t: TYPE
		do
			if name.to_string = as_none then
			else
				t := type_dictionary.reference_at(name)
				if t /= Void then
					lt := t.live_type
				end
				if lt /= Void then
					ct := lt.class_text
				else
					create cn.unknown_position(name)
					ct := class_text_dictionary.reference_at(cn)
				end
			end
			tfw.put_character('%N')
			if ct /= Void then
				tfw.put_string(once "class-path: %"")
				path := ct.path
				tfw.put_string(path)
				tfw.put_character('%"')
				tfw.put_character('%N')
				ct.id_extra_information(tfw)
			end
			if lt /= Void then
				lt.id_extra_information(tfw)
			end
		end

feature {CALL_PROC_CALL}
	covariance_check (call_site: POSITION; up_rf: RUN_FEATURE; run_time_set: RUN_TIME_SET) is
		require
			not call_site.is_unknown
			up_rf.arguments.count >= 1
			not run_time_set.is_empty
		local
			dyn_rf: RUN_FEATURE; up_args, dyn_args: FORMAL_ARG_LIST; r, a: INTEGER
			original_tm, redefined_tm: TYPE_MARK; original, redefined: TYPE
		do
			from
				r := 1
				up_args := up_rf.arguments
			until
				r > run_time_set.count
			loop
				dyn_rf := run_time_set.item(r).dynamic_feature(up_rf)
				from
					a := 1
					dyn_args := dyn_rf.arguments
				until
					a > up_args.count
				loop
					original_tm := up_args.type_mark(a)
					original := original_tm.resolve_in(up_rf.type_of_current)
					redefined_tm := dyn_args.type_mark(a)
					redefined := redefined_tm.resolve_in(dyn_rf.type_of_current)
					if not original.can_be_assigned_to(redefined) and then
						original.collected_feature_count > 0 and then
						original.collected_feature_count /= redefined.collected_feature_count
					 then
						error_handler.add_position(call_site)
						error_handler.append("Unsafe call site (see also next warning).")
						error_handler.print_as_warning						
						error_handler.append("Unsafe covariant redefinition of argument number ")
						error_handler.append_integer(a)
						error_handler.append(" (type %"")
						error_handler.append(original.name.to_string)
						error_handler.append("%" redefined as %"")
						error_handler.append(redefined.name.to_string)
						error_handler.append("%").")
						error_handler.add_position(up_rf.base_feature.start_position)
						error_handler.add_position(dyn_rf.base_feature.start_position)
						error_handler.print_as_warning
						status.set_safety_check_failed
					end
					a := a + 1
				end
				r := r + 1
			end
		end

feature {E_OLD}
	register_old (e_old: E_OLD) is
		require
			e_old /= Void
		do
			if old_list = Void then
				create old_list.with_capacity(2)
			end
			old_list.add_last(e_old)
		ensure
			old_list.fast_has(e_old)
		end

feature {}
	old_list: FAST_ARRAY[E_OLD]
			-- Non Void when some E_OLD have been gathered.

feature {RUN_FEATURE}
	get_and_clear_old_list: FAST_ARRAY[E_OLD] is
		do
			Result := old_list
			old_list := Void
		ensure
			old_list = Void
		end

feature {CALL_PROC_CALL, IMPLICIT_CAST}
	register_extra_local_expanded (type: TYPE) is
		require
			type.is_user_expanded
		do
			if extra_local_expanded_list = Void then
				create extra_local_expanded_list.with_capacity(4)
			end
			if not extra_local_expanded_list.fast_has(type) then
				extra_local_expanded_list.add_last(type)
			end
		ensure
			extra_local_expanded_list.fast_has(type)
		end

feature {}
	extra_local_expanded_list: FAST_ARRAY[TYPE]
			-- Non Void when some temporary variable is needed.
			--***PH(23/04/04) a single array can be reused once run_features
			--are removed.

feature {RUN_FEATURE, LIVE_TYPE}
	get_and_clear_extra_local_expanded_list: FAST_ARRAY[TYPE] is
		do
			Result := extra_local_expanded_list
			extra_local_expanded_list := Void
		ensure
			extra_local_expanded_list = Void
		end

feature {RUN_FEATURE, SWITCH, LIVE_TYPE, CLASS_INVARIANT, AGENT_CREATION, AGENT_ARGS}
	local_profile is
		require
			ace.profile
			cpp.on_c
		do
			cpp.put_string(once "se_local_profile_t local_profile;%N")
		end

	start_profile (rf: RUN_FEATURE) is
		require
			ace.profile
			cpp.on_c
			rf /= Void
		do
			cpp.put_string(once "local_profile.profile=profile+")
			cpp.put_integer(run_features.fast_index_of(rf))
			cpp.put_string(once ";%Nstart_profile(parent_profile, &local_profile);%N")
		end

	start_profile_class_invariant (t: LIVE_TYPE) is
		require
			ace.profile
			cpp.on_c
			t /= Void
		do
			register_class_invariant(t)
			cpp.put_string(once "local_profile.profile=inv_profile+")
			cpp.put_integer(class_invariants.fast_index_of(t))
			cpp.put_string(once ";%Nstart_profile(parent_profile, &local_profile);%N")
		end

	start_profile_agent_creation (ac: AGENT_CREATION) is
		require
			ace.profile
			cpp.on_c
			ac /= Void
		do
			register_agent_creation(ac)
			cpp.put_string(once "local_profile.profile=agent_profile+")
			cpp.put_integer(agent_creations.fast_index_of(ac))
			cpp.put_string(once ";%Nstart_profile(parent_profile, &local_profile);%N")
		end

	start_profile_agent_switch (t: TYPE) is
		require
			ace.profile
			cpp.on_c
			t /= Void
		do
			register_agent_switch(t)
			cpp.put_string(once "local_profile.profile=agent_switch_profile+")
			cpp.put_integer(agent_switches.fast_index_of(t))
			cpp.put_string(once ";%Nstart_profile(parent_profile, &local_profile);%N")
		end

	stop_profile is
		require
			ace.profile
			cpp.on_c
		do
			cpp.put_string(once "stop_profile(parent_profile, &local_profile);%N")
		end

feature {C_PRETTY_PRINTER}
	show_profile is
		require
			ace.profile
			cpp.on_c
		local
			n, i, s, a, sa: INTEGER
		do
			n := run_features.count
			if class_invariants /= Void then
				i := class_invariants.count
			end
			s := switch_collection.count
			if agent_creations /= Void then
				a := agent_creations.count
			end
			if agent_switches /= Void then
				sa := agent_switches.count
			end
			cpp.put_string(once "{%N%
                             %FILE *profile_file = fopen(%"profile.se%", %"w%");%N%
                             %if (profile_file!=NULL) {%N%
                             %int i;%N%
                             %se_profile_t sorted_profile[")
			cpp.put_integer(n)
			if i > 0 then
				cpp.put_string(once "];%Nse_profile_t sorted_inv_profile[")
				cpp.put_integer(i)
			end
			if s > 0 then
				cpp.put_string(once "];%Nse_profile_t sorted_switch_profile[")
				cpp.put_integer(s)
			end
			if a > 0 then
				cpp.put_string(once "];%Nse_profile_t sorted_agent_profile[")
				cpp.put_integer(a)
			end
			if sa > 0 then
				cpp.put_string(once "];%Nse_profile_t sorted_agent_switch_profile[")
				cpp.put_integer(sa)
			end
			cpp.put_string(once "];%Nse_profile_t sorted_all_profile[")
			cpp.put_integer(n+i+s+a+sa+3)

			-- sorting profiles

			cpp.put_string(once "];%Nmemcpy(sorted_profile, profile, ")
			cpp.put_integer(n)
			cpp.put_string(once "*sizeof(se_profile_t));%Nqsort(sorted_profile, ")
			cpp.put_integer(n)
			cpp.put_string(once ", sizeof(se_profile_t), profile_comparator);%N")
			cpp.put_string(once "memcpy(sorted_all_profile, profile, ")
			cpp.put_integer(n)
			cpp.put_string(once "*sizeof(se_profile_t));%N")
			if i > 0 then
				cpp.put_string(once "memcpy(sorted_inv_profile, inv_profile, ")
				cpp.put_integer(i)
				cpp.put_string(once "*sizeof(se_profile_t));%Nqsort(sorted_inv_profile, ")
				cpp.put_integer(i)
				cpp.put_string(once ", sizeof(se_profile_t), profile_comparator);%N")
				cpp.put_string(once "memcpy(sorted_all_profile+")
				cpp.put_integer(n)
				cpp.put_string(once ", inv_profile, ")
				cpp.put_integer(i)
				cpp.put_string(once "*sizeof(se_profile_t));%N")
			end
			if s > 0 then
				cpp.put_string(once "memcpy(sorted_switch_profile, switch_profile, ")
				cpp.put_integer(s)
				cpp.put_string(once "*sizeof(se_profile_t));%Nqsort(sorted_switch_profile, ")
				cpp.put_integer(s)
				cpp.put_string(once ", sizeof(se_profile_t), profile_comparator);%N")
				cpp.put_string(once "memcpy(sorted_all_profile+")
				cpp.put_integer(n+i)
				cpp.put_string(once ", switch_profile, ")
				cpp.put_integer(s)
				cpp.put_string(once "*sizeof(se_profile_t));%N")
			end
			if a > 0 then
				cpp.put_string(once "memcpy(sorted_agent_profile, agent_profile, ")
				cpp.put_integer(a)
				cpp.put_string(once "*sizeof(se_profile_t));%Nqsort(sorted_agent_profile, ")
				cpp.put_integer(a)
				cpp.put_string(once ", sizeof(se_profile_t), profile_comparator);%N")
				cpp.put_string(once "memcpy(sorted_all_profile+")
				cpp.put_integer(n+i+s)
				cpp.put_string(once ", agent_profile, ")
				cpp.put_integer(a)
				cpp.put_string(once "*sizeof(se_profile_t));%N")
			end
			if sa > 0 then
				cpp.put_string(once "memcpy(sorted_agent_switch_profile, agent_switch_profile, ")
				cpp.put_integer(sa)
				cpp.put_string(once "*sizeof(se_profile_t));%Nqsort(sorted_agent_switch_profile, ")
				cpp.put_integer(sa)
				cpp.put_string(once ", sizeof(se_profile_t), profile_comparator);%N")
				cpp.put_string(once "memcpy(sorted_all_profile+")
				cpp.put_integer(n+i+s+a)
				cpp.put_string(once ", agent_switch_profile, ")
				cpp.put_integer(sa)
				cpp.put_string(once "*sizeof(se_profile_t));%N")
			end
			cpp.put_string(once "sorted_all_profile[")
			cpp.put_integer(n+i+s+a+sa)
			cpp.put_string(once "]=root_profile;%N")
			cpp.put_string(once "sorted_all_profile[")
			cpp.put_integer(n+i+s+a+sa+1)
			cpp.put_string(once "]=runinit_profile;%N")
			cpp.put_string(once "sorted_all_profile[")
			cpp.put_integer(n+i+s+a+sa+2)
			cpp.put_string(once "]=atexit_profile;%N")
			cpp.put_string(once "qsort(sorted_all_profile, ")
			cpp.put_integer(n+i+s+a+sa+3)
			cpp.put_string(once ", sizeof(se_profile_t), profile_comparator);%N")

			-- summary

			cpp.put_string(once "fprintf(profile_file, %"===============================================================================\nSummary:\n%");%N")
			cpp.put_string(once "start_sumup_profile(profile_file);%Ni=")
			cpp.put_integer(n+i+s+a+sa+3)
			cpp.put_string(once ";while (i-->0) {%Nsumup_profile(profile_file, sorted_all_profile+i);%N}%N")

			-- detailed printing

			cpp.put_string(once "fprintf(profile_file, %"\n===============================================================================\nDetailed profile:\n%");%N")
			cpp.put_string(once "start_print_profile(profile_file);%N")
			cpp.put_string(once "print_profile(profile_file, &root_profile);%N")
			cpp.put_string(once "print_profile(profile_file, &runinit_profile);%N")
			cpp.put_string(once "print_profile(profile_file, &atexit_profile);%N")
			cpp.put_string(once "fprintf(profile_file, %"\n-------------------------------------------------------------------------------\n%");%Ni=")
			cpp.put_integer(n)
			cpp.put_string(once ";while (i-->0) {%Nprint_profile(profile_file, sorted_profile+i);%N}%N")
			if i > 0 then
				cpp.put_string(once "fprintf(profile_file, %"\n-------------------------------------------------------------------------------\n%");%Ni=")
				cpp.put_integer(i)
				cpp.put_string(once ";while (i-->0) {%Nprint_profile(profile_file, sorted_inv_profile+i);%N}%N")
			end
			if s > 0 then
				cpp.put_string(once "fprintf(profile_file, %"\n-------------------------------------------------------------------------------\n%");%Ni=")
				cpp.put_integer(s)
				cpp.put_string(once ";while (i-->0) {%Nprint_profile(profile_file, sorted_switch_profile+i);%N}%N")
			end
			if a > 0 then
				cpp.put_string(once "fprintf(profile_file, %"\n-------------------------------------------------------------------------------\n%");%Ni=")
				cpp.put_integer(a)
				cpp.put_string(once ";while (i-->0) {%Nprint_profile(profile_file, sorted_agent_profile+i);%N}%N")
			end
			if sa > 0 then
				cpp.put_string(once "fprintf(profile_file, %"\n-------------------------------------------------------------------------------\n%");%Ni=")
				cpp.put_integer(sa)
				cpp.put_string(once ";while (i-->0) {%Nprint_profile(profile_file, sorted_agent_switch_profile+i);%N}%N")
			end

			cpp.put_string(once "fprintf(profile_file, %"\n===============================================================================\n%");%N")
			cpp.put_string(once "fclose(profile_file);}%N")
			cpp.put_string(once "else fprintf(SE_ERR, %"Cannot open profile.se for writing.\n%");%N}%N")
		end

	define_extern_tables is
		require
			cpp.on_c
		local
			size: INTEGER
		do
			size := id_provider.max_id + 1
			cpp.macro_def(once "SE_MAXID", size)
			if generator_used then
				cpp.put_extern4(once "T7*g", size)
			end
			if generating_type_used then
				cpp.put_extern4(once "T7*t", size)
			end
			if ace.boost then
				if not gc_handler.is_off and then class_used(as_weak_reference) /= Void then
					initialize_size_table
				end
			else
				cpp.put_extern4(once "char*p", size)
				c_code.copy(once "void(*se_prinT[")
				size.append_in(c_code)
				c_code.append(once "])(FILE*,void**)")
				cpp.put_extern1(c_code)
				initialize_size_table
				if ace.sedb then
					c_code.copy(once "void*(*se_introspecT[")
					size.append_in(c_code)
					c_code.append(once "])(void*,char*,int*,int*)")
					cpp.put_extern1(c_code)
					c_code.copy(once "char* se_atT[")
					size.append_in(c_code)
					c_code.append(once "]")
					cpp.put_extern1(c_code)
				end
			end
			if ace.profile then
				cpp.put_extern4(once "se_profile_t profile", run_features.count)
				if switch_collection.count > 0 then
					cpp.put_extern4(once "se_profile_t switch_profile", switch_collection.count)
				end
				if agent_creations /= Void then
					cpp.put_extern4(once "se_profile_t agent_profile", agent_creations.count)
				end
				if agent_switches /= Void then
					cpp.put_extern4(once "se_profile_t agent_switch_profile", agent_switches.count)
				end
				if class_invariants /= Void then
					cpp.put_extern4(once "se_profile_t inv_profile", class_invariants.count)
				end
				cpp.put_extern1(once "se_profile_t root_profile")
				cpp.put_extern1(once "se_profile_t runinit_profile")
				cpp.put_extern1(once "se_profile_t atexit_profile")
				cpp.put_extern1(once "se_local_profile_t global_profile")
			end
		end

	initialize_size_table is
			-- initialize se_strucT
		require
			cpp.on_c
			not status.is_analyzing
		local
			i_ltm, i_c, nb_types: INTEGER; a_type: LIVE_TYPE; collection_sorter: COLLECTION_SORTER[LIVE_TYPE]
		do
			nb_types := id_provider.max_id + 1
			cpp.put_extern3(once "int se_strucT", nb_types)
			collection_sorter.sort(live_type_map)
			from
				i_ltm := 0
				i_c := 0
				c_code.clear_count
			until
				i_ltm > live_type_map.upper
			loop
				a_type := live_type_map.item(i_ltm)
				from
				until
					i_c >= a_type.id
				loop
					if c_code.count >= 100 then
						c_code.extend('%N')
						cpp.put_string(c_code)
						c_code.clear_count
					end
					c_code.append(once "0,")
					i_c := i_c + 1
				end
				check
					a_type.id = i_c
				end
				if c_code.count >= 100 then
					c_code.extend('%N')
					cpp.put_string(c_code)
					c_code.clear_count
				end
				if a_type.at_run_time then
					c_code.append(once "sizeof(T")
					i_c.append_in(c_code)
					c_code.append(once "),")
				else
					c_code.append(once "0,")
				end
				i_ltm := i_ltm + 1
				i_c := i_c + 1
			end
			-- last items are automatically zeroed by C compiler
			check
				c_code.last = ','
			end
			c_code.remove_last(1)
			c_code.append(once "};%N")
			cpp.put_string(c_code)
		ensure
			cpp.on_c
		end

	initialize_path_table is
		require
			ace.no_check
			cpp.on_c
		local
			i, id: INTEGER; ct: CLASS_TEXT; lt: LIVE_TYPE
		do
			from
				cpp.put_string(once "p[0]=%"???%";%N")
				i := 1
			until
				i > class_text_dictionary.count
			loop
				ct := class_text_dictionary.item(i)
				id := ct.id
				if id > 0 then
					cpp.put_string(once "p[")
					cpp.put_integer(id)
					cpp.put_string(once "]=")
					cpp.put_string_c(ct.path)
					cpp.put_string(fz_00)
				end
				i := i + 1
			end
			from
				i := live_type_map.upper
			until
				i < 0
			loop
				lt := live_type_map.item(i)
				if lt.at_run_time then
					cpp.put_string(once "se_prinT[")
					cpp.put_integer(lt.id)
					cpp.put_string(once "]=((void(*)(FILE*,void**))se_prinT")
					cpp.put_integer(lt.id)
					cpp.put_string(once ");%N")
					if lt.is_generic then
						cpp.put_string(once "p[")
						cpp.put_integer(lt.id)
						cpp.put_string(once "]=p[")
						cpp.put_integer(lt.class_text.id)
						cpp.put_string(once "];%N")
					end
					if ace.sedb then
						cpp.put_string(once "se_introspecT[")
						cpp.put_integer(lt.id)
						cpp.put_string(once "]=((void*(*)(void*,char*,int*,int*))se_introspecT")
						cpp.put_integer(lt.id)
						cpp.put_string(once ");%N")
						lt.prepare_introspection2
					end
				end
				i := i - 1
			end
		ensure
			cpp.on_c
		end

	initialize_generator is
		require
			cpp.on_c
		local
			i, id: INTEGER; ct: CLASS_TEXT; lt: LIVE_TYPE; n: STRING
		do
			from
				i := 1
			until
				i > class_text_dictionary.count
			loop
				ct := class_text_dictionary.item(i)
				id := ct.id
				if id >= 0 then
					cpp.put_array1('g', id)
					cpp.put_character('=')
					n := ct.name.to_string
					if n = as_integer then
						n := as_integer_32
					end
					cpp.put_se_string(n)
					cpp.put_string(fz_00)
				end
				i := i + 1
			end
			from
				i := live_type_map.upper
			until
				i < 0
			loop
				lt := live_type_map.item(i)
				if lt.at_run_time then
					ct := lt.class_text
					if ct.name.hashed_name /= lt.name then
						cpp.put_array1('g', lt.id)
						cpp.put_character('=')
						cpp.put_array1('g', ct.id)
						cpp.put_string(fz_00)
					end
				end
				i := i - 1
			end
		ensure
			cpp.on_c
		end

	initialize_generating_type is
		require
			cpp.on_c
		local
			i: INTEGER; lt: LIVE_TYPE; ct: CLASS_TEXT; rtm: STRING
		do
			from
				i := live_type_map.upper
			until
				i < 0
			loop
				lt := live_type_map.item(i)
				if lt.at_run_time then
					cpp.put_array1('t', lt.id)
					cpp.put_character('=')
					ct := lt.class_text
					rtm := lt.name.to_string
					if ct.name.to_string /= rtm then
						cpp.put_se_string(rtm)
					else
						cpp.put_array1('g', lt.id)
					end
					cpp.put_string(fz_00)
				end
				i := i - 1
			end
		ensure
			cpp.on_c
		end

	initialize_profile is
		require
			ace.profile
			cpp.on_c
		local
			i: INTEGER; b: STRING; rf: RUN_FEATURE; lt: LIVE_TYPE; t: TYPE; ac: AGENT_CREATION
		do
			from
				b := once "......................................."
				i := run_features.upper
			until
				i < run_features.lower
			loop
				rf := run_features.item(i)
				cpp.put_string(once "init_profile(profile+")
				cpp.put_integer(i)
				cpp.put_character(',')
				b.copy(rf.type_of_current.name.to_string)
				b.extend('.')
				rf.name.complete_name_in(b)
				cpp.put_string_c(b)
				cpp.put_string(fz_14)
				i := i - 1
			end
			if class_invariants /= Void then
				from
					b := once ""
					i := class_invariants.upper
				until
					i < class_invariants.lower
				loop
					lt := class_invariants.item(i)
					cpp.put_string(once "init_profile(inv_profile+")
					cpp.put_integer(i)
					cpp.put_character(',')
					b.copy(once "invariant of ")
					b.append(lt.name.to_string)
					cpp.put_string_c(b)
					cpp.put_string(fz_14)
					i := i - 1
				end
			end
			if agent_creations /= Void then
				from
					b := once ""
					i := agent_creations.upper
				until
					i < agent_creations.lower
				loop
					ac := agent_creations.item(i)
					cpp.put_string(once "init_profile(agent_profile+")
					cpp.put_integer(i)
					cpp.put_character(',')
					b.clear_count
					pretty_printer.expression_in(b, ac)
					cpp.put_string_c(b)
					cpp.put_string(fz_14)
					i := i - 1
				end
			end
			if agent_switches /= Void then
				from
					b := once ""
					i := agent_switches.upper
				until
					i < agent_switches.lower
				loop
					t := agent_switches.item(i)
					cpp.put_string(once "init_profile(agent_switch_profile+")
					cpp.put_integer(i)
					cpp.put_character(',')
					b.copy(once "agents ")
					b.append(t.canonical_type_mark.written_mark)
					cpp.put_string_c(b)
					cpp.put_string(fz_14)
					i := i - 1
				end
			end
			switch_collection.initialize_switch_site_profiles
		end

feature {RUN_FEATURE_3}
	register_for_c_define (rf: RUN_FEATURE_3) is
		require
			rf /= Void
		do
			registered_for_c_define.add_last(rf)
		end

feature {}
	collected_plug_in: SET[NATIVE_PLUG_IN] is
		once
			create {HASHED_SET[NATIVE_PLUG_IN]} Result.make
		end

feature {NATIVE_PLUG_IN}
	register_plug_in (native_plug_in: NATIVE_PLUG_IN) is
		require
			native_plug_in /= Void
		do
			collected_plug_in.add(native_plug_in)
		end

feature {LIVE_TYPE}
	set_deep_twin_used is
		do
			deep_twin_used := True
		end

	reference_to_separate (source_type: TYPE_MARK) is
			-- *** A REVOIR *** (Dom Oct 2004) ***
		require
			source_type.is_reference
		local
			i: INTEGER; lt: LIVE_TYPE; type: TYPE_MARK
		do
			from
				i := live_type_map.upper
			until
				i < 0
			loop
				lt := live_type_map.item(i)
				type := lt.canonical_type_mark
				if type.is_separate then
					if type.type.valid_redefinition_of(source_type.type) then
						type := type.actual_separate(source_type)
						lt := type.live_type
						if not lt.at_run_time then
							crash
							-- *** obsolete *** source_type.live_type.falling_down
							-- *** lt.set_at_run_time
						end
					else
						error_handler.cancel
					end
				end
				i := i - 1
			end
		end

feature {RUN_FEATURE_8, EXTERNAL_FUNCTION} --*** remove RUN_FEATURE_8
	set_generating_type_used is
		do
			generating_type_used := True
		end

	set_generator_used is
		do
			generator_used := True
		end

feature {CLUSTER}
	parse_include (include_name: STRING) is
		require
			include_name /= Void
		local
			ct: CLASS_TEXT
		do
			check
				parser_buffer.is_ready
			end
			echo.put_string(once "Handling include of %"")
			echo.put_string(include_name)
			echo.put_string(once "%" from ACE file. (Parsing %"")
			echo.put_string(parser_buffer.path)
			echo.put_string(once "%".)%N")
			ct := eiffel_parser.analyse_buffer
			check
				ct /= Void implies class_text_dictionary.has(ct.name)
			end
		end

feature {RUN_FEATURE}
	register_run_feature (rf: RUN_FEATURE) is
		require
			not_registered: not registered(rf)
		do
			if run_features = Void then
				create run_features.with_capacity(16)
			end
			run_features.add_last(rf)
		ensure
			run_features.last = rf
		end

feature {RUN_FEATURE, LIVE_TYPE}
	registered (rf: RUN_FEATURE): BOOLEAN is
		do
			Result := run_features /= Void and then run_features.fast_has(rf)
		end

feature {COMMAND_LINE_TOOLS}
	initialize_general_any_tuple is
			-- Some tools have to call this `initialize_general_any_tuple' once routine.
			--	Actually, `initialize_general_any_tuple' forces the creation of GENERAL, ANY and TUPLE first.
			-- Note, this is not in the creation of `smart_eiffel' itself because, not all tools are 
			-- supposed to load Eiffel classes.
		local
			hashed_string: HASHED_STRING; ct: CLASS_TEXT
		once
			-- Forcing first creation of GENERAL in order to initialize the machinery:
			hashed_string := string_aliaser.hashed_string(as_general)
			ct := class_text(create {CLASS_NAME}.unknown_position(hashed_string), True)

			-- Now forcing creation of ANY:
			hashed_string := string_aliaser.hashed_string(as_any)
			ct := class_text(create {CLASS_NAME}.make(hashed_string, ct.name.start_position), True)

			-- Finally, forcing creation of TUPLE:
			hashed_string := string_aliaser.hashed_string(as_tuple)
			ct := class_text(create {CLASS_NAME}.make(hashed_string, ct.name.start_position), True)

			-- Note: even the simple HELLO_WORLD, normally has to load TUPLE.
		end
	
feature {}
	register_class_invariant (t: LIVE_TYPE) is
		do
			if class_invariants = Void then
				create class_invariants.with_capacity(16)
			end
			if not class_invariants.fast_has(t) then
				class_invariants.add_last(t)
			end
		end

	register_agent_creation (ac: AGENT_CREATION) is
		do
			if agent_creations = Void then
				create agent_creations.with_capacity(16)
			end
			if not agent_creations.fast_has(ac) then
				agent_creations.add_last(ac)
			end
		end

	register_agent_switch (t: TYPE) is
		do
			if agent_switches = Void then
				create agent_switches.with_capacity(16)
			end
			if not agent_switches.fast_has(t) then
				agent_switches.add_last(t)
			end
		end

feature {}
	echo_magic_count (msg: STRING) is
		require
			msg /= Void
		do
			echo.put_string(msg)
			echo.put_string(once " (magic_count=")
			echo.put_integer(magic_count)
			echo.put_string(once ").%N")
			echo.flush
		end

	run_features: FAST_ARRAY[RUN_FEATURE]
			-- All the known features just prior the generation itself

	class_invariants: FAST_ARRAY[LIVE_TYPE]
			-- All the known types having an invariant during the generation

	agent_creations: FAST_ARRAY[AGENT_CREATION]
			-- All the known agent creations during the generation

	agent_switches: FAST_ARRAY[TYPE]
			-- All the known agent switches during the generation

	string_at_run_time: BOOLEAN is
			-- Pure query to know if the STRING type is used and is `at_run_time'?
		require
			is_ready
		local
			t: TYPE; hashed_string: HASHED_STRING; lt: LIVE_TYPE
		do
			hashed_string := string_aliaser.hashed_string(as_string)
			if hashed_string /= Void then
				t := type_dictionary.reference_at(hashed_string)
				if t /= Void then
					lt := t.live_type
				end
				if lt /= Void then
					Result := lt.at_run_time
				end
			end
		end

	get_started (root_class_name, root_procedure_name: STRING) is
			-- Get started to compile using creation procedure
			-- `root_procedure_name' of class text `root_class_name'.
		require
			not root_class_name.is_empty
			not root_procedure_name.is_empty
		local
			root_fn: FEATURE_NAME; root: CLASS_TEXT; root_type: TYPE; root_name: CLASS_NAME
			hashed_root_class_name: HASHED_STRING; i: INTEGER; lt: LIVE_TYPE
			fs: FEATURE_STAMP; af: ANONYMOUS_FEATURE
		do
			hashed_root_class_name := string_aliaser.hashed_string(root_class_name)
			if ace.no_check then
				set_generator_used
				set_generating_type_used
			end
			ace.parse_include
			create root_name.unknown_position(hashed_root_class_name)
			root := class_text(root_name, True)
			if root = Void then
				error_handler.append("Cannot load root class ")
				error_handler.append(root_class_name)
				error_handler.append(fz_dot_blank)
				error_handler.print_as_error
			elseif root.is_expanded then
				error_handler.add_position(root.name.start_position)
				error_handler.append("The root class must not be expanded (sorry, but this is a limitation of the compiler).")
				error_handler.print_as_fatal_error
			elseif root.is_generic then
				error_handler.append(root.name.to_string)
				error_handler.append(" cannot be a root class since it is a generic class.")
				error_handler.print_as_fatal_error
			else
				root_fn := root.root_creation_search(root_procedure_name)
				root_type := root.declaration_type_of_like_current
				if root_type.is_deferred then
					error_handler.append(root.name.to_string)
					error_handler.append(" cannot be a root class since it is a deferred class.")
					error_handler.print_as_warning
				end
				fs := root_type.lookup(root_fn)
				af := fs.anonymous_feature(root_type)
				if af.arguments /= Void then
					error_handler.add_position(af.start_position)
					error_handler.append("The main procedure must not have arguments.")
					error_handler.print_as_fatal_error
				end
			end
			if nb_errors = 0 then
				cecil_pool.parse_cecil_files
				collect_from_root(root_type, root_type.lookup(root_fn))
			end
			generic_formal_arguments_check
			if nb_errors = 0 and then not pretty_flag and then not short_or_class_check_flag then
				optimize(root_type, root_type.lookup(root_fn))
			end
			if nb_errors = 0 then
				status.set_adapting
				from
					i := live_type_map.upper
				until
					i < live_type_map.lower
				loop
					live_type_map.item(i).make_run_features
					i := i - 1
				end
				if nb_errors = 0 then
					from
						i := live_type_map.upper
					until
						i < live_type_map.lower
					loop
						lt := live_type_map.item(i)
						lt.adapt_run_features_and_class_invariant
						lt.make_switch_sites
						i := i - 1
					end
					if nb_errors = 0 then
						--***PH(03/03/04) Next lines may be removed I think.
						root_procedure ?= root_type.live_type.at(root_fn)
						check
							root_is_run_feature_3: root_procedure /= Void
						end
					end
				end
			end
			if nb_errors = 0 then
				safety_check
				echo.print_count(once "Loaded Classe", class_text_dictionary.count)
			end
		ensure
			nb_errors = 0 implies root_procedure /= Void
		end

	collect_from_root (root_type: TYPE; root_feature: FEATURE_STAMP) is
		local
			string_type, tmp: TYPE; i, magic: INTEGER
		do
			status.set_collecting
			assignment_handler.reset
			manifest_string_pool.reset
			manifest_array_pool.reset -- *** To be removed together with old << >> *** (Dom Oct 28th 2004) ***
			manifest_generic_pool.reset
			once_routine_pool.reset
			switch_collection.reset
			agent_pool.reset
			assignment_test_pool.reset
			--*** (PR 30/05/04) Reset is not yet complete. Should be
			--*** reseted: switches (waiting they are replaced by
			--*** inspects), manifest [unicode] strings,
			--***and maybe some other pools such as tuples, ...
			se_atexit_stamp := Void
			from
				i := live_type_map.upper
			until
				i < live_type_map.lower
			loop
				live_type_map.item(i).forget_previous_collect
				i := i - 1
			end
			echo_magic_count(once "Starting collect")
			-- Collect all the live features:
			if ace.no_check then
				-- STRINGs are needed for runtime support:
				string_type := type_string
				manifest_string_pool.collect_string(string_type)
			end
			from
				magic := -1 -- Iterate at least once
				-- Mark the root procedure as alive:
				tmp := collect(root_type, root_feature, True)
				check
					tmp = Void
				end
				-- Also mark possible -cecil features as alive:
				cecil_pool.collect
			until
				magic = magic_count
			loop
				magic := magic_count
				echo_magic_count(once "Before collect cycle")
				do_one_collect_cycle
			end
			status.set_collecting_done
			echo_magic_count(once "Collecting done")
			echo.put_string(once "Live_type_map size=")
			echo.put_integer(live_type_map.count)
			echo.put_new_line
		ensure
			status.collecting_done
		end

	do_one_collect_cycle is
		local
			i: INTEGER
		do
			from
				i := live_type_map.upper
			until
				i < live_type_map.lower
			loop
				live_type_map.item(i).propagate_features
				i := i - 1
			end
			from
				i := live_type_map.lower
			until
				i > live_type_map.upper
			loop
				live_type_map.item(i).do_collect
				i := i + 1 -- We go this way to collect new incoming live-types
			end
			assignment_handler.recompute_all_run_time_sets
		end

	compile_routines is
			-- Try to give the best order to the C output.
		require
			root_procedure.set_is_root
		local
			lt, lt_string: LIVE_TYPE; ct: TYPE_MARK; deep, i: INTEGER; stop: BOOLEAN; ctn: STRING
			cn, current_class_name: CLASS_NAME
		do
			if ace.no_check then
				echo.put_string(once "Compiling routines for ")
				echo.put_integer(live_type_map.count)
				echo.put_string(once " run classes :%N")
				cpp.swap_on_c
				from
					i := live_type_map.upper
				until
					i < 0
				loop
					lt := live_type_map.item(i)
					cn := lt.class_text_name
					if cn /= current_class_name then
						current_class_name := cn
					end
					lt.compile_to_c(0)
					i := i - 1
				end
			else
				echo.put_string(once "Compiling/Sorting ")
				echo.put_integer(live_type_map.count)
				echo.put_string(once " live TYPEs:%N")
				cpp.swap_on_c
				from
					i := live_type_map.upper
				until
					i < 0
				loop
					lt := live_type_map.item(i)
					ct := lt.canonical_type_mark
					if ct.is_kernel_expanded then
						lt.compile_to_c(0)
					elseif ct.is_string then
						lt_string := lt
					end
					i := i - 1
				end
				from
					i := live_type_map.upper
				until
					i < 0
				loop
					lt := live_type_map.item(i)
					i := i - 1
				end
				from
					i := live_type_map.upper
				until
					i < 0
				loop
					lt := live_type_map.item(i)
					ctn := lt.class_text_name.to_string
					if as_native_array = ctn then
						lt.compile_to_c(0)
					end
					i := i - 1
				end
				if lt_string /= Void then
					if lt_string.at_run_time then
						lt_string.compile_to_c(0)
					end
				end
				from
					i := live_type_map.upper
				until
					i < 0
				loop
					lt := live_type_map.item(i)
					ct := lt.canonical_type_mark
					ctn := ct.class_text_name.to_string
					if as_array = ctn or else as_fixed_array = ctn then
						lt.compile_to_c(0)
					end
					i := i - 1
				end
				from
					i := live_type_map.upper
				until
					i < 0
				loop
					lt := live_type_map.item(i)
					if lt.is_generic then
						lt.compile_to_c(0)
					end
					i := i - 1
				end
				from
					-- General sorting:
					deep := 6
				until
					stop
				loop
					from
						stop := True
						i := live_type_map.upper
					until
						i < 0
					loop
						lt := live_type_map.item(i)
						if not lt.compile_to_c_done then
							lt.compile_to_c(deep)
							if not lt.compile_to_c_done then
								stop := False
							end
						end
						i := i - 1
					end
					deep := deep - 1
					check
						deep >= -1
					end
				end
			end
		end

	prepare_introspection is
			-- Creates the introspection features
		require
			ace.no_check
		local
			i: INTEGER
		do
			cpp.swap_on_c
			from
				i := live_type_map.upper
			until
				i < 0
			loop
				live_type_map.item(i).prepare_introspection
				i := i - 1
			end
		end

	c_code: STRING is
		once
			create Result.make(128)
		end

	registered_for_c_define: FAST_ARRAY[RUN_FEATURE_3] is
		once
			create Result.with_capacity(512)
		end

	compile_registered_for_c_define is
		local
			i: INTEGER
		do
			from
				i := registered_for_c_define.upper
			until
				i < 0
			loop
				registered_for_c_define.item(i).c_define_or_scoop_define
				i := i - 1
			end
		end

	safety_check is
			-- Start final whole system analysis to decide whether this
			-- system is safe or not.
		require
			status.is_adapting
		local
			old_magic_count, i: INTEGER; do_it: BOOLEAN; lt: LIVE_TYPE
		do
			debug
				old_magic_count := magic_count
			end
			do_it := ace.safety_check
			status.set_safety_checking
			if do_it then
				echo_magic_count(once "Starting type safety check")
				from
					i := live_type_map.upper
				until
					i < live_type_map.lower
				loop
					lt := live_type_map.item(i)
					if lt.at_run_time then
						lt.safety_check
					end
					i := i - 1
				end
				status.set_safety_check_ok
			end
			debug
				check
					old_magic_count = magic_count
				end
			end
		ensure
			status.is_safety_checking
			ace.safety_check implies status.safety_check_done
		end

	optimize (root_type: TYPE; root_feature: FEATURE_STAMP) is
		require
			not pretty_flag
			not short_or_class_check_flag
			status.collecting_done
		local
			prev_magic_count: INTEGER
		do
			if ace.boost then
				echo.put_string(once "Starting optimization%N")
				if ace.boost1 then
					simplify
					collect_from_root(root_type, root_feature)
				elseif ace.boost2 then
					from
					until
						prev_magic_count = magic_count
					loop
						simplify
						prev_magic_count := magic_count
						--*** save run_time_sets
						--*** full collect collect
						--*** if run_time_sets = old run_time_sets then prev_magic_count := magic_count
					end
				else
					from
					until
						prev_magic_count = magic_count
					loop
						contextual_simplify
						prev_magic_count := magic_count
						--*** save run_time_sets
						--*** full collect collect
						--*** if run_time_sets = old run_time_sets then prev_magic_count := magic_count
					end
				end
				echo.put_string(once "Finished optimization (")
				echo.put_integer(magic_count)
				echo.put_string(once ")]%N")
			end
		end

	simplify is
		require
			not pretty_flag
			not short_or_class_check_flag
			status.collecting_done
			ace.boost1 or ace.boost2
		local
			lt: LIVE_TYPE; i: INTEGER; max_loop, prev_magic_count: INTEGER
		do
			echo_magic_count(once "Starting simplify")
			from
				max_loop := 50 -- (Is actually more than necessary, but avoids infinite inlinings.)
			until
				max_loop = 0 or else prev_magic_count = magic_count
			loop
				prev_magic_count := magic_count
				echo_magic_count(once "Before simplify cycle")
				from
					i := live_type_map.upper
				until
					i < 0
				loop
					lt := live_type_map.item(i)
					lt.simplify
					i := i - 1
				end
				max_loop := max_loop - 1
			end
			if max_loop = 0 then
				echo_magic_count(once "Symplify interrupted (infinite inlining ... or infinite recursion detected.).")
			end
			echo_magic_count(once "Symplify done")
		end

	contextual_simplify is
		require
			not pretty_flag
			not short_or_class_check_flag
			status.collecting_done
			ace.boost3
		local
			lt: LIVE_TYPE; i: INTEGER; prev_magic_count: INTEGER
		do
			echo.put_string(once "[Starting context_simplify (")
			echo.put_integer(magic_count)
			echo.put_string(once "): ")
			echo.flush
			from
			until
				prev_magic_count = magic_count
			loop
				prev_magic_count := magic_count
				echo.put_character('.')
				echo.flush
				from
					i := live_type_map.upper
				until
					i < 0
				loop
					lt := live_type_map.item(i)
					lt.contextual_simplify
					i := i - 1
				end
			end
			echo.put_string(once " finished context_simplify (")
			echo.put_integer(magic_count)
			echo.put_string(once ")]%N")
		end

	reference_integer_set: SET[LIVE_TYPE]

	singleton_memory: SMART_EIFFEL is
		once
			Result := Current
		end

invariant
	is_real_singleton: Current = singleton_memory

end -- class SMART_EIFFEL
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- SmartEiffel is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License,
-- as published by the Free Software Foundation; either version 2, or (at your option) any later version.
-- SmartEiffel is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have
-- received a copy of the GNU General Public License along with SmartEiffel; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
